/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.communication.jndi;

import java.util.Hashtable;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * InitialContextDecorator that wrap a real InitialContext object. All method
 * are delegated, except the close() method, which do nothing. This class allows
 * to shared an InitialContext to several users, preventing one of them to close
 * the context and block the other ones. Use the _close() method to close the
 * underlying InitialContext
 * 
 * @author alouis
 * 
 */
public class InitialContextDecorator extends InitialContext {

    /**
     * the jndi intial context.
     */
    private InitialContext initialContext;

    /**
     * Logging utilities.
     */
    private LoggingUtil log;

    /**
     * Create an InitialContext that wrap the given one. the close() method is
     * inactive.
     * 
     * @param ic
     *      the jndi initial context
     * @param log
     *      the logging util
     * @throws NamingException
     *      impossible to initialize the context
     */
    public InitialContextDecorator(final InitialContext ic, final LoggingUtil log)
            throws NamingException {
        this.initialContext = ic;
        this.log = log;
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.naming.InitialContext#addToEnvironment(java.lang.String,
     *  java.lang.Object)
     * @param propName
     *      the name
     * @param propVal
     *      the value
     * @throws NamingException
     *      impossible to add to environment
     * @return the associated object
     */
    @Override
    public Object addToEnvironment(final String propName, final Object propVal)
            throws NamingException {

        return this.initialContext.addToEnvironment(propName, propVal);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#bind(javax.naming.Name,
     * java.lang.Object)
     */
    @Override
    public void bind(Name name, Object obj) throws NamingException {
        this.log.call("Name:" + name + " - Object:" + obj);
        this.initialContext.bind(name, obj);

    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#bind(java.lang.String, java.lang.Object)
     */
    @Override
    public void bind(String name, Object obj) throws NamingException {
        this.log.call("String:" + name + " - Object:" + obj);
        this.initialContext.bind(name, obj);
    }

    /*
     * This InitialContext can not be closed by the user if it is seen as a pure
     * InitialContext. This method does not close the InitialContext
     */
    @Override
    public void close() throws NamingException {

        // Do not close the initialContext, use the specific one
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#composeName(javax.naming.Name,
     * javax.naming.Name)
     */
    @Override
    public Name composeName(Name name, Name prefix) throws NamingException {

        return this.initialContext.composeName(name, prefix);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#composeName(java.lang.String,
     * java.lang.String)
     */
    @Override
    public String composeName(String name, String prefix) throws NamingException {

        return this.initialContext.composeName(name, prefix);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#createSubcontext(javax.naming.Name)
     */
    @Override
    public Context createSubcontext(Name name) throws NamingException {
        this.log.call("Name:" + name);

        return this.createSubcontext(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#createSubcontext(java.lang.String)
     */
    @Override
    public Context createSubcontext(String name) throws NamingException {
        this.log.call("String:" + name);

        return this.initialContext.createSubcontext(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#destroySubcontext(javax.naming.Name)
     */
    @Override
    public void destroySubcontext(Name name) throws NamingException {
        this.log.call("Name:" + name);

        this.initialContext.destroySubcontext(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#destroySubcontext(java.lang.String)
     */
    @Override
    public void destroySubcontext(String name) throws NamingException {
        this.log.call("String:" + name);

        this.initialContext.destroySubcontext(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#getEnvironment()
     */
    @Override
    public Hashtable<?, ?> getEnvironment() throws NamingException {
        return this.initialContext.getEnvironment();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#getNameInNamespace()
     */
    @Override
    public String getNameInNamespace() throws NamingException {

        return this.initialContext.getNameInNamespace();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#getNameParser(javax.naming.Name)
     */
    @Override
    public NameParser getNameParser(Name name) throws NamingException {

        return this.initialContext.getNameParser(name);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#getNameParser(java.lang.String)
     */
    @Override
    public NameParser getNameParser(String name) throws NamingException {

        return this.initialContext.getNameParser(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#list(javax.naming.Name)
     */
    @Override
    public NamingEnumeration<NameClassPair> list(Name name) throws NamingException {
        log.call("Name:" + name);

        NamingEnumeration<NameClassPair> res = this.initialContext.list(name);

        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#list(java.lang.String)
     */
    @SuppressWarnings("unchecked")
    @Override
    public NamingEnumeration<NameClassPair> list(String name) throws NamingException {
        log.call("String:" + name);

        NamingEnumeration<NameClassPair> res = this.initialContext.list(name);

        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#listBindings(javax.naming.Name)
     */
    @SuppressWarnings("unchecked")
    @Override
    public NamingEnumeration<Binding> listBindings(Name name) throws NamingException {
        log.call("Name:" + name);

        NamingEnumeration<Binding> res = this.initialContext.listBindings(name);

        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#listBindings(java.lang.String)
     */
    @SuppressWarnings("unchecked")
    @Override
    public NamingEnumeration<Binding> listBindings(String name) throws NamingException {
        log.call("String:" + name);

        NamingEnumeration<Binding> res = this.initialContext.listBindings(name);

        return res;
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.naming.InitialContext#lookup(javax.naming.Name)
     */
    @Override
    public Object lookup(final Name name) throws NamingException {

        Object res = this.lookup(name.toString());

        return res;
    }

    /**
     * (non-Javadoc).
     * 
     * 
     * @see javax.naming.InitialContext#lookup(java.lang.String)
     */
    @Override
    public Object lookup(final String name) throws NamingException {
        log.call("String:" + name);

        Object res = null;
        String methodName = "lookup;";
        if ((SystemUtil.isCacheActivated()) && (name.contains("@"))) {
            this.log.debug("Get from cache");
            res = JNDIEndpointCache.getInstance().getEntry(methodName + name);
            if (res == null) {
                res = this.initialContext.lookup(name);
                if (res != null) {
                    this.log.debug("Put in cache");
                    JNDIEndpointCache.getInstance().putEntry(methodName + name, res);
                }
            }
        } else {
            res = this.initialContext.lookup(name);
        }

        return res;
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.naming.InitialContext#lookupLink(javax.naming.Name)
     */
    @Override
    public Object lookupLink(Name name) throws NamingException {

        Object res = this.lookupLink(name.toString());

        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#lookupLink(java.lang.String)
     */
    @Override
    public Object lookupLink(String name) throws NamingException {
        log.call("String:" + name);

        Object res = null;
        String methodName = "lookupLink;";
        if ((SystemUtil.isCacheActivated()) && (name.contains("@"))) {
            this.log.debug("Get from cache");
            res = JNDIEndpointCache.getInstance().getEntry(methodName + name);
            if (res == null) {
                res = this.initialContext.lookupLink(name);
                if (res != null) {
                    this.log.debug("Put in cache");
                    JNDIEndpointCache.getInstance().putEntry(methodName + name, res);
                }
            }
        } else {
            res = this.initialContext.lookupLink(name);
        }

        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#rebind(javax.naming.Name,
     * java.lang.Object)
     */
    @Override
    public void rebind(Name name, Object obj) throws NamingException {
        log.call("Name:" + name + " - Object:" + obj);
        
        this.initialContext.rebind(name, obj);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#rebind(java.lang.String,
     * java.lang.Object)
     */
    @Override
    public void rebind(String name, Object obj) throws NamingException {
        log.call("String:" + name + " - Object:" + obj);
        
        this.initialContext.rebind(name, obj);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#removeFromEnvironment(java.lang.String)
     */
    @Override
    public Object removeFromEnvironment(String propName) throws NamingException {

        return this.initialContext.removeFromEnvironment(propName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#rename(javax.naming.Name,
     * javax.naming.Name)
     */
    @Override
    public void rename(Name oldName, Name newName) throws NamingException {
        log.call("OldName:" + oldName + " - NewName:" + newName);

        this.initialContext.rename(oldName, newName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#rename(java.lang.String,
     * java.lang.String)
     */
    @Override
    public void rename(String oldName, String newName) throws NamingException {
        log.call("OldString:" + oldName + " - NewString:" + newName);
        
        this.initialContext.rename(oldName, newName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#unbind(javax.naming.Name)
     */
    @Override
    public void unbind(Name name) throws NamingException {
        log.call("Name:" + name);
        
        this.initialContext.unbind(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.naming.InitialContext#unbind(java.lang.String)
     */
    @Override
    public void unbind(String name) throws NamingException {
        log.call("String:" + name);
        
        this.initialContext.unbind(name);
    }

    /**
     * Close the underlying InitialContext wrapped object.
     * 
     * @throws NamingException
     */
    public void closeInitialContextDecorator() throws NamingException {
        this.initialContext.close();
    }

}
