/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.communication.jndi;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Fractal component used to retrieve the JNDI initial context.
 * 
 * @author alouis, Christophe HAMERLING - eBM WebSourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.communication.jndi.JNDIService.class))
public class JNDIServiceImpl implements JNDIService {

    /**
     * Logging utilities.
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Users context
     */
    protected Context usersContext;

    private Hashtable<String, Object> jndiEnv;

    private InitialContextDecorator initialContextDecorator;

    @LifeCycle(on = LifeCycleType.START)
    public void start() {
        log = new LoggingUtil(logger);
        jndiEnv = new Hashtable<String, Object>(PropertyUtil.loadServerProperties());
        jndiEnv.put(JAVA_NAMING_FACTORY_LOGGER, this.log);
        
        log.debug("JNDI properties used : " + this.jndiEnv);
    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() {
        if (initialContextDecorator != null) {
            try {
                log.debug("Close the JNDI InitialContext");
                initialContextDecorator.closeInitialContextDecorator();
            } catch (NamingException e) {
                log.error(e);
            } finally {
                initialContextDecorator = null;
                log = null;
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.kernel.jndi.JNDIService#getJNDIContext()
     */
    public InitialContext getInitialContext() throws NamingException {
        log.call();
        
        if (initialContextDecorator == null) {
            log.debug("Create a new wrapped InitialContext");
            InitialContext ic = new InitialContext(this.jndiEnv);
            initialContextDecorator = new InitialContextDecorator(ic, log);
        }
        
        return initialContextDecorator;
    }

    public Context getUsersContext() throws NamingException {
        log.call();
        
        if (usersContext == null) {
            try {
                usersContext = (Context) getInitialContext().lookup(USERS_REF);
            } catch (NameNotFoundException e) {
                try {
                    usersContext = getInitialContext().createSubcontext(USERS_REF);
                } catch (final NameAlreadyBoundException e1) {
                    // In a PEtAL network this context can be propagated by JNDI
                    // engine between the lookup and the creation
                    usersContext = (Context) getInitialContext().lookup(USERS_REF);
                }
            } catch (NameAlreadyBoundException e) {
                // in distributed mode, it is possible that 2 containers are
                // creating the context in the same time, the second
                // container get an Already Bound exception
                usersContext = (Context) getInitialContext().lookup(USERS_REF);
            }
        }
        return usersContext;
    }
}
