/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: TribeThread.java 11:51:28 AM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.communication.jndi.tribe.thread;

import java.util.Map;

import javax.naming.NamingException;

import org.objectweb.petals.communication.jndi.tribe.TribeServer;
import org.objectweb.petals.communication.jndi.tribe.msg.request.NewMasterRequest;
import org.objectweb.petals.communication.jndi.tribe.msg.request.RegistryRequest;
import org.objectweb.petals.communication.jndi.tribe.msg.response.ExceptionResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.FullUpdateResponse;
import org.objectweb.petals.communication.jndi.tribe.msg.response.RegistryResponse;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This class is the Tribe listener for the JNDI directory.
 * 
 * Each time we received a message, we need to check if the received message is
 * not a message sent by this server 
 * 
 * Overwise we need to check if the recipient is :
 * <ul>
 * <li>0 for everybody except the sender of the message</li>
 * <li>a none-zero value for only one recipient</li>
 * </ul>
 * 
 * Each thread is associated with one JNDI server. This thread is used to
 * communicate with the other tribe based JNDI servers.
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class TribeThread extends Thread {

    /**
     * Last request date
     */
    protected static long lastRequest;

    /**
     * Last response date
     */
    protected static long lastResponse;

    /**
     * Logger
     */
    protected LoggingUtil log;

    /**
     * Boolean used to control the lifecycle of the Tribe listener
     */
    protected boolean run = true;

    /**
     * Id of the sender
     */
    protected long sender;

    /**
     * JNDI server
     */
    protected TribeServer server;

    /**
     * Creates a new instance of {@link TribeThread}
     * 
     * @param server
     */
    public TribeThread(TribeServer server) {
        super("TribeThread" + server.getStartTime());
        this.server = server;
        this.log = server.getLog();
        this.sender = server.getStartTime();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    public void run() {
        log.call();

        while (run) {
            Object msg = null;
            RegistryRequest request = null;
            RegistryResponse response = null;
            try {
                // wait for an incoming message on the multicast channel
                // TODO need to block until message is completely processed
                msg = server.getChannel().receive();
                
                if (msg instanceof RegistryRequest) {
                    request = (RegistryRequest) msg;
                    if (request.getSender() != sender
                        && (request.getRecipient() == sender || request
                            .getRecipient() == 0)
                        && lastRequest != request.getRequestDate()) {

                        // process request
                        lastRequest = request.getRequestDate();
                        processRequest(request);
                    }
                } else if (msg instanceof RegistryResponse) {
                    response = (RegistryResponse) msg;
                    if (response.getSender() != sender
                        && (response.getRecipient() == sender || response
                            .getRecipient() == 0)
                        && lastResponse != response.getResponseDate()) {

                        // process response
                        lastResponse = response.getResponseDate();
                        processResponse(response);
                    }
                } else {
                    if (msg != null) {
                        log.warning("Message is not a registry message "
                            + msg.getClass());
                    }
                }
            } catch (Exception e) {
                log.error(e.getMessage(), e);
            }
        }
    }

    /**
     * Shutdown the Tribe listener
     * 
     */
    public void shutdown() {
        log.call();
        run = false;
    }

    /**
     * Process a request
     * 
     * @param request
     *            request to process
     */
    protected void processRequest(RegistryRequest request) {
        RegistryResponse response = null;
        try {
            Object object = server.processRequest(request);
            switch (request.getType()) {
                /*
                 * BIND
                 */
                case bind:
                    break;
                /*
                 * CREATE SUBCONTEXT
                 */
                case createSubcontext:
                    break;
                /*
                 * DESTROY SUBCONTEXT
                 */
                case destroySubcontext:
                    break;
                /*
                 * FULL UPDATE
                 */
                case fullUpdate:
                    response = new FullUpdateResponse((Map<?, ?>) object,
                        sender, request.getSender());
                    break;
                /*
                 * REBIND
                 */
                case rebind:
                    break;
                /*
                 * RENAME
                 */
                case rename:
                    break;
                /*
                 * UNBIND
                 */
                case unbind:
                    break;
                default:
                    break;
            }
        } catch (NamingException e) {
            response = new ExceptionResponse(e, sender, request.getSender());
        }
        if (response != null) {
            // We send back the response
            try {
                server.getChannel().send(response);
            } catch (Exception e) {
                log.error("Problem while sending response", e);
            }
        }
    }

    /**
     * Process a response
     * 
     * @param response
     *            response to process
     */
    protected void processResponse(RegistryResponse response) {
        RegistryRequest request = null;
        Object outObject = server.processResponse(response);

        switch (response.getType()) {
            /*
             * MY START TIME
             */
            case myStartTime:
                if (outObject != null) {
                    // This server can be a master
                    request = new NewMasterRequest(sender, response.getSender());
                }
                break;
            default:
                break;
        }

        if (request != null) {
            // We send back the request
            try {
                server.getChannel().send(request);
            } catch (Exception e) {
                log.error("Problem while sending request", e);
            }
        }
    }
}
