/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.communication.network;

import java.util.List;
import java.util.Map;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.network.ContainerInformation.ContainerState;

/**
 * The interface of the container network service. The container network is used
 * to get container informations {@link ContainerInformation} of the system
 * containers.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public interface NetworkService {

    enum Status {
        NOT_INITIALIZED, STARTED, STOPPED
    }

    /**
     * Delete the container context from the JNDI server.
     *
     * @throws PetalsException
     */
    public void deleteContainerContext() throws PetalsException;

    /**
     * Test if the container is started. Get the container information from the
     * JNDI server and its state.
     *
     * @param containerName
     * @return
     * @throws PetalsException
     */
    public boolean isContainerStarted(String containerName)
        throws PetalsException;

    /**
     * Test if the container is registered.
     *
     * @param containerName
     * @return true if the container is registered
     */
    public boolean isContainerRegistered(String containerName);

    /**
     * Register a new container in the JNDI server. The container will be
     * rebinded if it already exists.
     *
     * @param container
     * @throws PetalsException
     */
    public void registerContainer(ContainerInformation container)
        throws PetalsException;

    /**
     * Unregister the container from the JNDI server.
     *
     * @param containerName
     * @throws PetalsException
     */
    public void unregisterContainer(String containerName)
        throws PetalsException;
    
    /**
     * Update the local view against the remote container state.
     *
     * @param state
     * @throws PetalsException
     */
    public void updateContainerState(String containerName, ContainerState state)
        throws PetalsException;

    /**
     * Update the local container state.
     *
     * @param state
     * @throws PetalsException
     */
    public void updateLocalContainerState(ContainerState state)
        throws PetalsException;

    /**
     * Retrieve the container information.
     *
     * @param containerName
     * @return the container information or null if it has not been found
     * @throws PetalsException
     */
    public ContainerInformation getContainerInformation(String containerName)
        throws PetalsException;

    /**
     * Get all the containers information. If the state is null,
     * returns all the containers information.
     *
     * @return a list of {@link ContainerInformation}
     */
    public List<ContainerInformation> retrieveAllContainersInformation(ContainerInformation.ContainerState state);

    /**
     * Get the local container information
     * @return the local {@link ContainerInformation}
     */
    public ContainerInformation getLocalContainerInformation();

    /**
     * Get all the container names with the given state. If the state is null,
     * returns all the containers.
     *
     * @param state
     * @return
     */
    public List<String> getAllContainersNames(
        ContainerInformation.ContainerState state);
    
    /**
     * Get the local view of the domain.
     */
    public Map<String, ContainerState> getNetworkView();
    
    /**
     * Refresh the local view of the domain.
     */
    public void refreshNetworkView();

    /**
     * Get the status of the container network
     *
     * @return
     */
    public Status getStatus();


}
