/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.communication.network;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.communication.network.ContainerInformation.ContainerState;
import org.objectweb.petals.kernel.configuration.ConfigurationService;
import org.objectweb.petals.util.JNDIUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This {@link NetworkService} implementation is based on JNDI. It uses the JNDI
 * registry that <b>MUST</b> be started before.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.communication.network.NetworkService.class))
public class NetworkServiceImpl implements NetworkService {

    private Map<String, ContainerState> containerStatesMap;

    /**
     * Name of the JNDI container context
     */
    static final String CONTAINERS_REF = "containers";

    /**
     * Logging utilities
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "jndi", signature = org.objectweb.petals.communication.jndi.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * The local container name
     */
    private String localContainerName;

    /**
     * The status of the current network service
     */
    private Status status = Status.NOT_INITIALIZED;

    /**
     * The container topology mode : true if TRIBe based, false if JORAM based.
     */
    private boolean dynamic = false;

    /**
     * The timer for the ping service.
     */
    private Timer timer;

    /**
     * The initial context.
     */
    private Context rootContext;

    /**
     * The container context.
     */
    private Context containersContext;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#createContainerContext
     * ()
     */
    private Context getOrCreateContainerContext() throws PetalsException {
        log.start();

        try {
            rootContext = jndiService.getInitialContext();
            //dirty trick due to synchronization issue with the JORAM JNDI startup
            try {
                this.containersContext = (Context) rootContext.lookup(CONTAINERS_REF);
            } catch (NameNotFoundException e) {
                try {
                    this.containersContext = rootContext.createSubcontext(CONTAINERS_REF);
                } catch (NameAlreadyBoundException e1) {
                    this.containersContext = (Context) rootContext.lookup(CONTAINERS_REF);
                }
            }
        } catch (NamingException e) {
            throw new PetalsException(
                    "Problem while trying to initialize the container JNDI context: "
                            + e.getMessage(), e);
        } finally {
            log.end();
        }
        return containersContext;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#deleteContainerContext
     * ()
     */
    public void deleteContainerContext() throws PetalsException {
        log.start();
        try {
            if (JNDIUtil.isBound(rootContext, "", CONTAINERS_REF)) {
                rootContext.unbind(CONTAINERS_REF);
            }
        } catch (NamingException e) {
            throw new PetalsException("Problem while trying to unregister the containers context",
                    e);
        }
        log.end();
    }

    /**
     * Get all the containers name with the given state. If the state is null,
     * returns all the container names.
     * 
     * @return
     */
    public List<String> getAllContainersNames(ContainerInformation.ContainerState state) {

        log.start();

        List<String> result = new ArrayList<String>();
        List<ContainerInformation> containers = retrieveAllContainersInformation(state);
        for (ContainerInformation information : containers) {
            if (state == null) {
                result.add(information.getName());
            } else {
                if (state == ContainerInformation.ContainerState.STARTED
                        && information.getState() == ContainerInformation.ContainerState.STARTED) {
                    result.add(information.getName());
                } else if (state == ContainerInformation.ContainerState.STOPPED
                        && information.getState() == ContainerInformation.ContainerState.STOPPED) {
                    result.add(information.getName());
                }
            }
        }
        log.end();
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.petals.communication.network.NetworkService#
     * getContainerInformation(java.lang.String)
     */
    public ContainerInformation getContainerInformation(String containerName)
            throws PetalsException {
        log.start("Get container information : " + containerName);
        ContainerInformation info = null;

        try {
            info = (ContainerInformation) this.containersContext.lookup(containerName);
        } catch (NamingException e) {
            log.debug("Container information has not been found for container : " + containerName);
        }

        log.end("Returns : " + info);
        return info;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.petals.communication.network.NetworkService#
     * getLocalContainerInformation()
     */
    public ContainerInformation getLocalContainerInformation() {
        ContainerInformation info = null;
        try {
            info = getContainerInformation(localContainerName);
        } catch (PetalsException e) {
            log.debug("Can not get container information from JNDI");
        }

        return info;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#isContainerStarted
     * (java.lang.String)
     */
    public boolean isContainerStarted(String containerName) throws PetalsException {
        log.start("Container started ? " + containerName);
        if (containerName == null) {
            throw new IllegalArgumentException("The container name must not be null");
        }

        boolean isStarted = false;

        if (!this.containerStatesMap.containsKey(containerName)) {
            // the local view may have not been yet updated against the
            // container start-up
            ContainerInformation containerInformation = this.getContainerInformation(containerName);
            this.containerStatesMap.put(containerName, containerInformation.getState());
        }

        if (this.containerStatesMap.get(containerName) == ContainerState.STARTED) {
            isStarted = true;
        }

        log.end("Returns : " + isStarted);
        return isStarted;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#isContainerRegistered
     * (java.lang.String)
     */
    public boolean isContainerRegistered(String containerName) {
        log.start("Container registered ? " + containerName);

        if (containerName == null) {
            throw new IllegalArgumentException("The container name must not be null");
        }

        boolean isRegistered = false;

        if (JNDIUtil.isBound(rootContext, CONTAINERS_REF, containerName)) {
            isRegistered = true;
        }

        log.end("Returns : " + isRegistered);
        return isRegistered;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#registerContainer
     * (org.objectweb.petals.kernel.admin.ContainerInformation)
     */
    public void registerContainer(ContainerInformation containerInformation) throws PetalsException {
        log.start("Register container : " + containerInformation.getName());

        try {
            if (!JNDIUtil.isBound(rootContext, CONTAINERS_REF, containerInformation.getName())) {
                this.containersContext.bind(containerInformation.getName(), containerInformation);
            } else {
                this.containersContext.rebind(containerInformation.getName(), containerInformation);
            }
        } catch (NamingException e) {
            throw new PetalsException("Error while binding container information", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.petals.communication.network.NetworkService#
     * updateContainerState(java.lang.String,
     * org.objectweb.petals.communication.
     * network.ContainerInformation.ContainerState)
     */
    public void updateContainerState(String containerName, ContainerState state)
            throws PetalsException {
        log.call();
        this.containerStatesMap.put(containerName, state);

    }

    /*
     * (non-Javadoc)
     * 
     * org.objectweb.petals.kernel.admin.ContainerInformation.ContainerState)
     */
    public void updateLocalContainerState(ContainerState state) throws PetalsException {
        log.start("Update container to state  " + state);

        try {
            if (JNDIUtil.isBound(rootContext, CONTAINERS_REF, localContainerName)) {
                ContainerInformation containerInformation = (ContainerInformation) containersContext
                        .lookup(localContainerName);
                containerInformation.setState(state);
                containersContext.rebind(localContainerName, containerInformation);
            }
        } catch (NamingException e) {
            log.error("Problem while updating container state", e);
        }
        log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.petals.communication.network.NetworkService#
     * retrieveAllContainersInformation()
     */
    public List<ContainerInformation> retrieveAllContainersInformation(
            ContainerInformation.ContainerState state) {
        log.start();

        List<ContainerInformation> containers = new ArrayList<ContainerInformation>();
        try {
            NamingEnumeration<Binding> enums = rootContext.listBindings(CONTAINERS_REF);
            while (enums.hasMoreElements()) {
                Binding binding = enums.next();
                ContainerInformation containerInformation = (ContainerInformation) binding
                        .getObject();

                this.containerStatesMap.put(containerInformation.getName(), containerInformation
                        .getState());
                if (state == null) {
                    containers.add(containerInformation);
                } else {
                    if (state == ContainerInformation.ContainerState.STARTED
                            && containerInformation.getState() == ContainerInformation.ContainerState.STARTED) {
                        containers.add(containerInformation);
                    } else if (state == ContainerInformation.ContainerState.STOPPED
                            && containerInformation.getState() == ContainerInformation.ContainerState.STOPPED) {
                        containers.add(containerInformation);
                    }
                }
            }
        } catch (NamingException e) {
            log
                    .warning("Problem while trying to retrieve all the containers information : context may not be initialized");
        }

        log.end("Returns " + containers.size() + " containers");
        return containers;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#unregisterContainer
     * (java.lang.String)
     */
    public void unregisterContainer(String containerName) throws PetalsException {
        log.start("Unregistering container : " + containerName);

        try {
            if (JNDIUtil.isBound(rootContext, CONTAINERS_REF, containerName)) {
                this.containersContext.unbind(containerName);
            }
        } catch (NamingException e) {
            throw new PetalsException("Container unregistration failure", e);
        }

        log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.communication.network.NetworkService#refreshNetworkView
     * ()
     */
    public void refreshNetworkView() {
        PingTask pingTask = new PingTask(this, log, this.containerStatesMap);
        pingTask.run();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.communication.network.NetworkService#getNetworkView
     * ()
     */
    public Map<String, ContainerState> getNetworkView() {
        return this.containerStatesMap;
    }

    /**
     * Start Fractal component
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        log = new LoggingUtil(logger);
        log.start();
        try {
            dynamic = SystemUtil.isDynamicTopology();
            containerStatesMap = new ConcurrentHashMap<String, ContainerState>();

            startComponent();

            int pingPeriod;

            try {
                pingPeriod = Integer.parseInt(SystemUtil.getPingPeriod());
            } catch (NumberFormatException e) {
                throw new NumberFormatException("A ping period has a wrong format : "
                        + e.getMessage());
            }

            this.timer = new Timer();
            this.timer.schedule(new PingTask(this, log, this.containerStatesMap), 1000, pingPeriod);
        } catch (final Exception e) {
            this.log.error("An error occurs during NetworkService starting", e);
            throw e;
        } finally {
            log.end();
        }
    }

    /**
     * Stop Fractal component
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        log.start();
        try {
            stopComponent();
        } catch (Exception e) {
            log.error("Can not stop the container network : " + e.getMessage());
            throw e;
        }

        if (this.timer != null) {
            this.timer.cancel();
        }

        log.end();
    }

    /**
     * Retrieve the initial container information. In static mode, simply get
     * the default container information from the server properties. In
     * distributed mode, get the information from the already registered
     * containers and generate all unusued values if necessary.
     */
    private ContainerInformation retrieveInitialContainerInformation() throws PetalsException {
        ContainerInformation result = null;
        if (dynamic) {
            result = retrieveDynamicContainerInformation();
        } else {
            result = retrieveStaticContainerInformation();
        }
        return result;
    }

    /**
     * Get a dynamic configuration which doesn't interfer with others
     * containers. Usefull when multiple containers are started on the same
     * host. The new container can not be started on the same ports.
     * 
     * @return a container information object with unused ports
     */
    private ContainerInformation retrieveDynamicContainerInformation() {

        log.start();
        ContainerInformation containerInformation = getDefaultDynamicContainerInformation();

        // calculate the dynamic configuration only at the intialisation of the
        // ContainerNetworkService
        if (status == Status.NOT_INITIALIZED) {
            // get the container information from the registered containers
            List<ContainerInformation> allContainersInformation = retrieveAllContainersInformation(null);

            // get new configuration values from the other containers
            ConfigurationService config = new ConfigurationService(containerInformation,
                    allContainersInformation, log);
            containerInformation = config.setUpConfiguration();
        }

        log.end();
        return containerInformation;
    }

    /**
     * 
     * @return
     */
    private ContainerInformation retrieveStaticContainerInformation() {
        log.call();
        return getDefaultStaticContainerInformation();
    }

    /**
     * Creates the initial container context in the JNDI registry. The JNDI
     * registry must be started to do this.
     * 
     * 
     * @throws PetalsException
     */
    private void startComponent() throws PetalsException {
        if (status != Status.NOT_INITIALIZED) {
            throw new PetalsException("Can not initialize component in this state : " + getStatus());
        }
        getOrCreateContainerContext();

        ContainerInformation localContainerInformation = retrieveInitialContainerInformation();

        registerContainer(localContainerInformation);
        localContainerName = localContainerInformation.getName();

        status = Status.STARTED;
    }

    /**
     * Stop the component.
     * 
     * @throws PetalsException
     */
    private void stopComponent() throws PetalsException {
        if (status != Status.STARTED) {
            throw new PetalsException("Can not stop component in this state : " + getStatus());
        }

        status = Status.STOPPED;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.communication.network.ContainerNetworkService#getStatus
     * ()
     */
    public Status getStatus() {
        return status;
    }

    /**
     * Get the default container information for a dynamic configuration. Set
     * only the settings from the server properties or the default values for
     * static information. If a value is null, it would be calculated
     * dynamically.
     * 
     * @return
     */
    private static ContainerInformation getDefaultDynamicContainerInformation() {

        // Set global values
        ContainerInformation containerInformation = new ContainerInformation();
        containerInformation.setName(SystemUtil.getValue(SystemUtil.CONTAINER_NAME));
        containerInformation.setState(ContainerInformation.ContainerState.UNKNOWN);
        containerInformation.setHost(SystemUtil.getHost());
        containerInformation.setUid(SystemUtil.getContainerUID());

        // Set JORAM properties
        containerInformation.setJoramId(SystemUtil.getValue(SystemUtil.JORAM_ID));
        containerInformation.setJoramDomainPort(SystemUtil.getValue(SystemUtil.JORAM_DOMAIN_PORT));
        containerInformation.setJoramLogin(SystemUtil.getJoramUser());
        containerInformation.setJoramPassword(SystemUtil.getJoramPassword());
        containerInformation.setJoramTCPPort(SystemUtil.getValue(SystemUtil.JORAM_TCP_PORT));
        containerInformation.setJoramDomain(SystemUtil.getJoramDomain());

        // Set DREAM properties
        containerInformation.setDreamTCPPort(SystemUtil.getValue(SystemUtil.DREAM_TCP_PORT));

        // set JMX properties
        containerInformation.setJmxJndiPort(SystemUtil.getValue(SystemUtil.JMX_JNDI_PORT));
        containerInformation.setJmxPort(SystemUtil.getValue(SystemUtil.JMX_PORT));
        containerInformation.setJmxLogin(SystemUtil.getValue(SystemUtil.JMX_USER));
        containerInformation.setJmxPassword(SystemUtil.getValue(SystemUtil.JMX_PWD));

        // Set HTML properties
        containerInformation.setHtmlPort(SystemUtil.getValue(SystemUtil.HTML_PORT));

        // Set JNDI properties
        containerInformation.setJndiServer(SystemUtil.getJNDIServer());
        containerInformation.setJndiFactory(SystemUtil.getJndiFactory());
        containerInformation.setJndiPort(SystemUtil.getJndiPort());

        // Set Extra properties
        containerInformation.setProperties(PropertyUtil.loadPropertiesFile(PropertyUtil.SERVER_PROPS));

        return containerInformation;
    }

    /**
     * Get the default container information for a static configuration. Set the
     * settings from the server properties or the default ones.
     * 
     * @return
     */
    private static ContainerInformation getDefaultStaticContainerInformation() {

        // Set global values
        ContainerInformation containerInformation = new ContainerInformation();
        containerInformation.setName(SystemUtil.getContainerName());
        containerInformation.setState(ContainerInformation.ContainerState.UNKNOWN);
        containerInformation.setHost(SystemUtil.getHost());

        // Set JORAM properties
        containerInformation.setJoramId(SystemUtil.getJoramId());
        containerInformation.setJoramDomainPort(SystemUtil.getJoramDomainPort());
        containerInformation.setJoramLogin(SystemUtil.getJoramUser());
        containerInformation.setJoramPassword(SystemUtil.getJoramPassword());
        containerInformation.setJoramTCPPort(SystemUtil.getJoramTCPPort());
        containerInformation.setJoramDomain(SystemUtil.getJoramDomain());

        // Set DREAM properties
        containerInformation.setDreamTCPPort(SystemUtil.getDreamTCPPort());

        // set JMX properties
        containerInformation.setJmxJndiPort(SystemUtil.getJmxJndiPort());
        containerInformation.setJmxPort(SystemUtil.getJmxPort());
        containerInformation.setJmxLogin(SystemUtil.getJmxUser());
        containerInformation.setJmxPassword(SystemUtil.getJmxPassword());

        // Set HTML properties
        containerInformation.setHtmlPort(SystemUtil.getHtmlPort());

        // Set JNDI properties
        containerInformation.setJndiServer(SystemUtil.getJNDIServer());
        containerInformation.setJndiFactory(SystemUtil.getJndiFactory());
        containerInformation.setJndiPort(SystemUtil.getJndiPort());
        containerInformation.setUid(SystemUtil.getContainerUID());

        // Set Extra properties
        containerInformation.setProperties(PropertyUtil.loadPropertiesFile(PropertyUtil.SERVER_PROPS));

        return containerInformation;
    }

}
