/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationServiceMBean.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.container.lifecycle;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.management.ObjectName;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.container.event.StateChangeFailedEvent;
import org.ow2.petals.jbi.descriptor.ServiceAssembly;
import org.objectweb.petals.platform.systemstate.SystemStateService;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This class manage the life cycle of deployed service assemblies
 *
 * @author Olivier Fabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
public class ServiceAssemblyLifeCycle extends LifeCycleAbstract {

    private URI installationRoot;

    /**
     * Platform component :: SystemState Service
     */
    @Requires(name = "systemstate", signature = org.objectweb.petals.platform.systemstate.SystemStateService.class)
    private SystemStateService recoverySrv;

    /**
     * The component
     */
    @Service
    private Component comp;

    private ServiceAssembly serviceAssembly;

    private Map<String, ServiceUnitLifeCycle> serviceUnitsLifeCycles;

    public ServiceAssemblyLifeCycle() {
        super();
    }

    /**
     * Accessor to the fractal component
     *
     * @return the fractal component
     */
    public Component getComponent() {
        return comp;
    }

    /**
     *
     * @param mbeanName
     *            name of the MBean for this LifeCycle. In fact, the
     *            DeploymentServiceMBean name is used as there is no
     *            SALifeCycleMBean
     * @param log
     * @param serviceAssembly
     * @param repositoryService
     * @param systemStateService
     * @param installationRoot
     */
    public void init(ObjectName mbeanName, LoggingUtil log,
            ServiceAssembly serviceAssembly, URI installationRoot) {
        super.init(mbeanName, log);

        this.installationRoot = installationRoot;
        this.serviceAssembly = serviceAssembly;
        this.serviceUnitsLifeCycles = Collections
                .synchronizedMap(new LinkedHashMap<String, ServiceUnitLifeCycle>());
    }

    /**
     * Do nothing
     */
    @Override
    public void doInit() throws JBIException {

    }

    @Override
    public void doShutdown() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            List<ServiceUnitLifeCycle> suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(
                    sus.values());
            Collections.reverse(suLifeCycles);
            for (Iterator iter = suLifeCycles.iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.shutDown();
            }
        }
    }

    @Override
    public void doStart() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            for (Iterator iter = sus.values().iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.start();
            }
        }
    }

    @Override
    public void doStop() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this
                .getServiceUnitsLifeCycles();

        synchronized (sus) {
            List<ServiceUnitLifeCycle> suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(
                    sus.values());
            Collections.reverse(suLifeCycles);
            for (Iterator iter = suLifeCycles.iterator(); iter.hasNext();) {
                ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter
                        .next();
                lifeCycle.stop();
            }
        }
    }

    public URI getInstallationRoot() {
        return installationRoot;
    }

    public ServiceAssembly getServiceAssembly() {
        return serviceAssembly;
    }

    /**
     * Return the synchronized map of deployed service units.
     *
     * @return
     */
    public Map<String, ServiceUnitLifeCycle> getServiceUnitsLifeCycles() {
        return serviceUnitsLifeCycles;
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////

    /**
     * Register a ServiceUnitLifeCycle into a synchronized map of the
     * ServiceAssemblyLifeCycle
     *
     * @param suName
     *            name of a service unit
     * @param serviceUnitLifeCycle
     *            life cycle of the service unit
     * @throws JBIException
     */
    public void registerSUIntoSA(String suName,
            ServiceUnitLifeCycle serviceUnitLifeCycle) throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            if (!sus.containsKey(suName)) {
                sus.put(suName, serviceUnitLifeCycle);
            } else {
                throw new JBIException("Duplicate service unit : same name");
            }

        }
    }

    /**
     * Unregister a ServiceUnitLifeCycle from a synchronized map of the
     * ServiceAssemblyLifeCycle
     *
     * @param suName
     *            name of a service unit
     */
    public void unregisterSUFromSA(String suName) {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            sus.remove(suName);
        }
    }

    @Override
    public void setState(String state) throws JBIException {
        super.setState(state);
        String saName = this.serviceAssembly.getIdentification().getName();
        try {
            recoverySrv.updateServiceAssemblyState(saName, state);
        } catch (Exception e) {
            String msg = "Service assembly state can't be persisted for recovery";
            log.error(msg, e);
            throw new JBIException(msg, e);
        }
    }

    /**
     * No init state for ServiceAssemblyLifeCycle
     */
    @Override
    @LifeCycle(on = LifeCycleType.START)
    public void start() throws JBIException {
        log.start();

        if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    setState(STARTING);

                    doStart();

                    setState(STARTED);
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "A runtime exception occured during the start.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be started in this state: "
                    + state;

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

    /**
     * Stop of the petals component
     *
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stopComponentLifeCycle() {
        log.call();
    }
}
