/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DeploySUToComponentask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.task.deployment.deploy;

import java.net.URI;
import java.util.HashMap;

import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.management.LifeCycleMBean;

import org.objectweb.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.descriptor.ServiceUnit;
import org.objectweb.petals.jbi.management.ManagementException;
import org.objectweb.petals.jbi.management.admin.AdminServiceMBean;
import org.objectweb.petals.jbi.management.task.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.task.deployment.DeploymentUtils;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This task deploys service unit to the targeted component <br/> Inputs : SU
 * {@link JBIDescriptor}, SU install root URI <br/> Outputs : XML result for of
 * the deployment as {@link String} <br/>
 *
 * @author ofabre - EBM Websourcing
 *
 */
public class DeploySUToComponentask implements Task {

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected AdminServiceMBean adminService;

    public DeploySUToComponentask(LoggingUtil log,
        AdminServiceMBean adminService) {
        super();
        this.log = log;
        this.adminService = adminService;
    }

    @SuppressWarnings("unchecked")
    public void execute(HashMap context) throws Exception {

        URI installationRoot = (URI) context
            .get(DeploymentContextConstants.SU_INSTALL_ROOT);

        ServiceUnit serviceUnit = (ServiceUnit) context
            .get(DeploymentContextConstants.SU_DESCRIPTOR);

        String xmlResult = deploySUtoComponent(installationRoot, serviceUnit);

        /*
         * Fill context
         */
        context.put(DeploymentContextConstants.COMPONENT_XML_RESULT, xmlResult);

    }

    /**
     * Deploy the given service unit to the targeted component. The component
     * must be in the STARTED state to allow deployment of service units.
     *
     * @param installationRoot
     *            the su installation root {@link URI}
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit deployment on the
     *         targeted component (see section 6.9 of the JSR 208)
     * @throws ManagementException
     *             if component and related object are missing, if the call of
     *             the service unit manager method "deploy" fails or if the
     *             component isn't in the STARTED state
     */
    protected String deploySUtoComponent(URI installationRoot,
        ServiceUnit serviceUnit) throws ManagementException {

        /*
         * Retrieve the target component life cycle
         */
        String targetedComponent = DeploymentUtils
            .getServiceUnitTargetedComponent(serviceUnit);
        ComponentLifeCycle clc = adminService
            .getComponentLifeCycleByName(targetedComponent);
        if (clc == null) {
            String msg = "The service unit can not be deployed, as the component '"
                + targetedComponent + "' is not installed.";
            log.error(msg);
            throw new ManagementException(msg);
        }

        /*
         * The component must be in the STARTED state to allow deployment of
         * service units
         */
        if (!LifeCycleMBean.STARTED.equals(clc.getCurrentState())) {
            String msg = "Component "
                + targetedComponent
                + " isn't started : you can't deploy service units on a not started component !";
            log.error(msg);
            throw new ManagementException(msg);
        }

        /*
         * Retrieve the service unit manager of the component that allow to
         * process deployment
         */
        ServiceUnitManager sum = clc.getComponent().getServiceUnitManager();
        if (sum == null) {
            String msg = "Component "
                + targetedComponent
                + " doesn't support deployment : service unit manager is null !";
            log.error(msg);
            throw new ManagementException(msg);
        }

        String xmlReturn = null;
        try {
            xmlReturn = sum.deploy(DeploymentUtils
                .getServiceUnitName(serviceUnit), installationRoot.getPath());
        } catch (DeploymentException e) {
            String msg = "Deployement of the service unit "
                + DeploymentUtils.getServiceUnitName(serviceUnit)
                + " has failed (caused by a component \"" + targetedComponent
                + "\" su manager exception).";
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }
        return xmlReturn;
    }

    @SuppressWarnings("unchecked")
    public void undo(HashMap context) {
        log.debug("Revert DeploySUToComponent");
        URI installationRoot = (URI) context
            .get(DeploymentContextConstants.SU_INSTALL_ROOT);

        ServiceUnit serviceUnit = (ServiceUnit) context
            .get(DeploymentContextConstants.SU_DESCRIPTOR);

        try {
            undeploySUFromComponent(installationRoot, serviceUnit);
        } catch (ManagementException e) {
            String msg = "Failed to revert a DeploySUToComponentask";
            log.error(msg, e);
        }
    }

    /**
     * Undeploy the given service unit from the targeted component. The
     * component must be in the STARTED state to allow deployment of service
     * units.
     *
     * @param installationRoot
     *            the su installation root {@link URI}
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit deployment on the
     *         targeted component (see section 6.9 of the JSR 208)
     * @throws ManagementException
     *             if component and related object are missing, if the call of
     *             the service unit manager method "undeploy" fails or if the
     *             component isn't in the STARTED state
     */
    public String undeploySUFromComponent(URI installationRoot,
        ServiceUnit serviceUnit) throws ManagementException {
        /*
         * Retrieve the target component life cycle
         */
        String targetedComponent = DeploymentUtils
            .getServiceUnitTargetedComponent(serviceUnit);
        ComponentLifeCycle clc = adminService
            .getComponentLifeCycleByName(targetedComponent);
        if (clc == null) {
            String msg = "The component lifecycle element for service unit deployment must be non null.";
            log.error(msg);
            throw new ManagementException(msg);
        }

        /*
         * The component must be in the STARTED state to allow deployment of
         * service units
         */
        if (!LifeCycleMBean.STARTED.equals(clc.getCurrentState())) {
            String msg = "Component "
                + targetedComponent
                + " isn't started : you can not deploy a service unit on a component that has not been started.";
            log.error(msg);
            throw new ManagementException(msg);
        }

        /*
         * Retrieve the service unit manager of the component that allow to
         * process deployment
         */
        ServiceUnitManager sum = clc.getComponent().getServiceUnitManager();
        if (sum == null) {
            String msg = "Component "
                + targetedComponent
                + " doesn't support deployment : service unit manager is null !";
            log.error(msg);
            throw new ManagementException(msg);
        }

        String xmlReturn = null;
        String serviceUnitName = DeploymentUtils
                        .getServiceUnitName(serviceUnit);
        try {
            xmlReturn = sum.undeploy(serviceUnitName, installationRoot.getPath());
        } catch (DeploymentException e) {
            String msg = "Undeployement of the service unit "
                + serviceUnitName
                + " has failed (caused by a targeted component \""
                + serviceUnitName
                + "\" su manager exception).";
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }
        return xmlReturn;
    }
}
