/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: EndpointRegistry.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.registry;



import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.naming.NamingException;
import javax.xml.namespace.QName;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jmx.DistributedJMXServer;
import org.objectweb.petals.communication.jmx.JMXService;
import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.jbi.management.admin.AdminServiceMBean;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.ConsumerEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.InternalEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.LinkedEndpoint;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.ParameterCheckHelper;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.petals.util.WSDLUtil;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.commons.threadlocal.DocumentBuilders;
import org.ow2.petals.commons.threadlocal.Transformers;
import org.w3c.dom.Document;

/**
 * This class is used to manage everything that deals with endpoint in the
 * distributed environment
 * 
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.jbi.messaging.registry.EndpointService.class))
public class EndpointServiceImpl implements EndpointService {
    
    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name = "jmx", signature = org.objectweb.petals.communication.jmx.JMXService.class)
    protected JMXService jmxService;

    /**
     * JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.objectweb.petals.communication.jndi.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Endpoints cache fractal component
     */
    @Requires(name = "endpoint-cache", signature = EndpointCacheService.class)
    protected EndpointCacheService endpointCache;

    /**
     * Admin service fractal component
     */
    @Requires(name = "admin", signature = org.objectweb.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * JNDI Registry instance
     */
    protected JNDIRegistry registry;

    /**
     * The local container name.
     */
    private String containerName;

    /**
     * Thread pool to update remote new-endpoints asycnhronously: minimum number
     * of threads
     */
    // TODO: change the thread pool implementation to be able to have no idle
    // thread.
    private static int MIN_NEWENDPOINTS_UPDATER = 1;

    /**
     * Thread pool to update remote new-endpoints asycnhronously: maximum number
     * of threads
     */
    // TODO: fix this value according to the number of PEtAL nodes in the
    // network
    private static int MAX_NEWENDPOINTS_UPDATER = 50;

    /**
     * Thread pool to update remote new-endpoints asycnhronously
     */
    private ThreadPoolExecutor threadPoolRemoteNewEndpointsUpdater;
    

    /**
     * The timer to update the new endpoints in background
     */
    private Timer updateLocalNewEndpointsTimer;

    private AtomicBoolean updateScheduled;

    /**
     * Private scheduled task which update the local new endpoints in
     * background.
     * 
     * @author rnaudin
     * 
     */
    private class UpdateLocalNewEndpointsTimerTask extends TimerTask {

        @Override
        public void run() {
            EndpointServiceImpl.this.log.start();
            try {
                // We update only local new-endpoints
                while (EndpointServiceImpl.this.updateNewEndpoints(true) > 0) {
                    // The processing is included in the condition
                }
            } catch (JBIException e) {
                EndpointServiceImpl.this.log.error(e);
            }
            
            EndpointServiceImpl.this.updateScheduled.set(false);
            EndpointServiceImpl.this.log.end();
        }

    }

    /**
     * @see EndpointService#activateEndpoint(QName, String, ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName, String endpointName,
            ConsumerEndpoint address) throws JBIException {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        if (address == null) {
            throw new IllegalArgumentException("Consumer endpoint must be non null");
        }
        
        try {
        	final InternalEndpoint se = new InternalEndpoint(serviceName, endpointName, address
	                .getComponentName(), address.getContainerName(), this);
	        this.registry.registerInternalEndpoint(se);
	
	        if (updateScheduled.compareAndSet(false, true)) {
	            this.updateLocalNewEndpointsTimer.schedule(new UpdateLocalNewEndpointsTimerTask(), 5000);
	        }
	        return se;
	        
        } finally {
        	this.log.end();
        }
    }

    /**
     * @see EndpointService#createConnection(QName, QName, String)
     */
    public void createConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();

        if (consInterface == null) {
            throw new IllegalArgumentException("Interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);

        this.log.end();
    }

    /**
     * @see EndpointService#createConnection(QName, String, QName, String)
     */
    public void createConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();

        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);

        this.log.end();
    }

    /**
     * @see EndpointService#deactivateEndpoint(ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterInternalEndpoint((AbstractEndpoint) endpoint);

        this.log.end();
    }

    /**
     * @see EndpointService#deleteConnection(QName, QName, String)
     */
    public void deleteConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Consumer interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointService#deleteConnection(QName, String, QName, String)
     */
    public void deleteConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointService#deregisterExternalEndpoint(ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterExternalEndpoint(endpoint);
        this.log.end();
    }

    /**
     * @see EndpointService#getEndpoint(QName, String) <ol>
     *  <li>Retreive the named endpoint</li>
     *  <li>If found, update the endpoint service of the found endpoint with
     *  this</li>
     *  </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        this.log.call();
        return this.getEndpoint(service, name, true);
    }

    /**
     * @see EndpointService#getEndpoint(QName, String,boolean) <ol>
     *  <li>Retrieve the named endpoint</li>
     *  <li>If found, update the endpoint service of the found endpoint with
     *  this</li>
     *  </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name, boolean resolveLink) {
        this.log.start();
        if (service == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        AbstractEndpoint abstractEndpoint = null;
        try {
            abstractEndpoint = this.registry.getInternalEndpoint(service, name, resolveLink);
        } catch (RegistryException e) {
            // no endpoint found, return null;
        }

        this.log.end();
        return abstractEndpoint;
    }

    /**
     * @see EndpointService#getEndpointDescriptorForEndpoint(ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;

        if ((endpoint instanceof AbstractEndpoint)) {
            document = ((AbstractEndpoint) endpoint).getDescription();
            if (document == null) {
                try {
                    // update the all the new endpoints if necessary
                    if (isNewEndpoints(endpoint)) {
                        this.updateNewEndpoints(false);
                    }

                    // retrieve the endpoint
                    AbstractEndpoint abstractEndpoint = this.registry.getInternalEndpoint(endpoint
                            .getServiceName(), endpoint.getEndpointName());

                    document = abstractEndpoint.getDescription();
                    // set both description and interfaces
                    ((AbstractEndpoint) endpoint).setDescription(document);
                    ((AbstractEndpoint) endpoint).setInterfaces(abstractEndpoint.getInterfaces());
                } catch (RegistryException e) {
                    this.log.error("Error while validating endpoint " + endpoint.getEndpointName()
                            + " for service " + endpoint.getServiceName(), e);
                }
            }
        }
        this.log.end();
        return document;
    }

    /**
     * @see EndpointService#getExternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(QName interfaceName) {
        this.log.start();

        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.registry.getExternalEndpointsForInterface(interfaceName);
        } catch (RegistryException e) {
            this.log.error("Error while getting external endpoints for interface " + interfaceName,
                    e);
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointService#getExternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        endpoints = this.registry.getExternalEndpointsForService(serviceName);

        this.log.end();
        return endpoints;
    }

    /**
     * Return the interface for the specified endpoint
     * 
     * @param endpoint
     *      endpoint on which we want the interface names; must be non null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }

        // update only the first time
        if (endpoint.getInterfaces() == null) {
            try {
                if (isNewEndpoints(endpoint)) {
                    this.updateNewEndpoints(false);
                }

                AbstractEndpoint abstractEndpoint = this.registry.getInternalEndpoint(endpoint
                        .getServiceName(), endpoint.getEndpointName());
                // set both description and interfaces
                endpoint.setInterfaces(abstractEndpoint.getInterfaces());
                endpoint.setDescription(abstractEndpoint.getDescription());
            } catch (Exception e) {
                this.log.error("Error while getting interfaces for endpoint "
                        + endpoint.getEndpointName() + " of service " + endpoint.getServiceName(),
                        e);
            }
        }

        this.log.end();
        return endpoint.getInterfaces();
    }

    /**
     * @see EndpointService#getInternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(final QName interfaceName) {
        this.log.start();
        
        ServiceEndpoint[] endpoints = null;
        if (SystemUtil.isCacheActivated()) {
        	endpoints = this.endpointCache.getInterfaceEndpoints(interfaceName);
        }
        if (endpoints == null) {
	        endpoints = new ServiceEndpoint[0];
	        try {
	            // Update the new endpoints
	            this.updateNewEndpoints(false);
	
	            endpoints = this.registry.getInternalEndpointsForInterface(interfaceName);
	            
	            if (SystemUtil.isCacheActivated() && endpoints.length > 0) {
	            	this.endpointCache.putInterfaceEndpoints(interfaceName, endpoints);
	            }
	            
	        } catch (RegistryException e) {
	            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
	                    e);
	        } catch (JBIException e) {
	            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
	                    e);
	        }
        }

        this.log.end();
        return endpoints;
        
    }

    /**
     * @see EndpointService#getInternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(final QName serviceName) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        
        ServiceEndpoint[] endpoints = null;
        if (SystemUtil.isCacheActivated()) {
        	endpoints = this.endpointCache.getServiceEndpoints(serviceName);
        }
        if (endpoints == null) {
        	endpoints = this.registry.getInternalEndpointsForService(serviceName);

            if (SystemUtil.isCacheActivated() && endpoints.length > 0) {
            	this.endpointCache.putServiceEndpoints(serviceName, endpoints);
            }
        }

        this.log.end();
        return endpoints;
    }

    /**
     * @see
     *  EndpointService#isExchangeWithConsumerOkayForComponent(InternalEndpoint
     *  , MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint, "Endpoint", this.log);
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = this.jmxService
                    .createDistributedJMXServer(internalEndpoint.getContainerName());
            // Ask the component if it is ok to accept the exchange
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_CONS, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see
     *  EndpointService#isExchangeWithProviderOkayForComponent(InternalEndpoint
     *  , MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = this.jmxService
                    .createDistributedJMXServer(internalEndpoint.getContainerName());
            // Retrieve the endpoint description
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_PROV, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see EndpointService#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.registerExternalEndpoint(externalEndpoint);
        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        this.log = new LoggingUtil(this.logger);
        this.log.call();
        
        this.threadPoolRemoteNewEndpointsUpdater = new ThreadPoolExecutor(
                EndpointServiceImpl.MIN_NEWENDPOINTS_UPDATER,
                EndpointServiceImpl.MAX_NEWENDPOINTS_UPDATER,
                60, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>());

        this.registry = new JNDIRegistry(this, this.jndiService, this.log);
        this.containerName = SystemUtil.getContainerName();
        this.updateLocalNewEndpointsTimer = new Timer(true);
        this.updateScheduled = new AtomicBoolean(false);

        try {
            this.registry.initContexts();
        } catch (RegistryException e) {
            throw new IllegalLifeCycleException(e.toString());
        }

    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.call();
        this.registry = null;
        
        this.threadPoolRemoteNewEndpointsUpdater.shutdownNow();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getDescription
     * (java.lang.String, java.lang.String)
     */
    public Document getDescription(String serviceName, String endpointName) throws JBIException {
        this.log.start();

        ServiceEndpoint serviceEndpoint = this
                .getEndpoint(QName.valueOf(serviceName), endpointName);

        this.log.end();
        return this.getEndpointDescriptorForEndpoint(serviceEndpoint);

    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllExternalEndpoints()
     */
    public List<Map<String, Object>> getAllExternalEndpoints() {
        this.log.start();

        List<Map<String, Object>> externalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] externalEndpoints = this.getExternalEndpointsForInterface(null);

        for (ServiceEndpoint serviceEndpoint : externalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            endpointMap.put(KEY_TYPE, TYPE_EXTERNAL);
            externalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return externalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllInternalEndpoints()
     */
    public List<Map<String, Object>> getAllInternalEndpoints() {
        this.log.start();

        List<Map<String, Object>> internalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] internalEndpoints = this.getInternalEndpointsForInterface(null);

        for (ServiceEndpoint serviceEndpoint : internalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName().toString());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            if (serviceEndpoint instanceof LinkedEndpoint) {
                LinkedEndpoint linkEP = (LinkedEndpoint) serviceEndpoint;
                endpointMap.put(KEY_TYPE, linkEP.getToServiceName().toString() + "@"
                        + linkEP.getToEndpointName());
            } else {
                endpointMap.put(KEY_TYPE, TYPE_INTERNAL);
            }
            internalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return internalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getAllEndpoints
     * ()
     */
    public List<Map<String, Object>> getAllEndpoints() {
        this.log.call();

        List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
        allEndpointsMap.addAll(this.getAllInternalEndpoints());

        return allEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * updateNewEndpoints(boolean)
     */
    public long updateNewEndpoints(boolean remoteContainerInvocation) throws JBIException {
        this.log.start();

        long updatedLocalNewEndpoints = 0;
        int remoteTaskNumber = 0;
        final List<AbstractEndpoint> newLocalEndpoints = new ArrayList<AbstractEndpoint>();
        final List<String> containersToUpdate = new ArrayList<String>();
        final CompletionService<Long> ecs = new ExecutorCompletionService<Long>(
                threadPoolRemoteNewEndpointsUpdater);

        // Synchronize this part if a remote and a local invocation are running
        // concurrently
        synchronized (this.containerName) {

            // First, a list of the new local endpoints and a list of remote
            // containers containing new-endpoints are build
            this.registry.retrieveNewEndpoints(newLocalEndpoints, containersToUpdate);

            // Next, several tasks will be processed asynchronously. One task
            // for the processing of local endpoints, and one task per remote
            // container containing new-endpoints. As the local-new endpoint
            // updates must be synchronized to prevent remote and a local
            // invocations, the remote new-endpoint update tasks are submit
            // during processing of local new-endpoints
    
            final Future<Long> updateLocalNewEndointsFuture;
            if (newLocalEndpoints.size() > 0) {
                // Submit the task updating local endpoint description and interface in the registry
                final Callable<Long> updateEndpointDescriptionAndInterfacesTask = new Callable<Long>() {
                    public Long call() {
                        EndpointServiceImpl.this.log.debug("Updating local new-endpoints...");
    
                        long updatedNewEndpoints = 0;
                        // Update the local new endpoints
                        for (final AbstractEndpoint abstractEndpoint : newLocalEndpoints) {
                            EndpointServiceImpl.this.updateEndpointDescriptionAndInterfaces(abstractEndpoint);
                            try {
                                EndpointServiceImpl.this.registry.validateEndpoint(abstractEndpoint);
                                updatedNewEndpoints++;
                            } catch (final JBIException e) {
                                EndpointServiceImpl.this.log.error("Problem while updating new endpoint '"
                                        + abstractEndpoint.getEndpointName() + "'", e);
                            }
                        }
    
                        EndpointServiceImpl.this.log.debug(updatedNewEndpoints + " local new-endpoints updated.");
                        return Long.valueOf(updatedNewEndpoints);    
                    }
                };
                updateLocalNewEndointsFuture = ecs.submit(updateEndpointDescriptionAndInterfacesTask);
                this.log.debug("Updating local new-endpoints task added to the pool.");
            }
            else {
                updateLocalNewEndointsFuture = null;
                this.log.debug("No local new-endpoints to update -> No task added to the pool.");
            }

            // Invoke update of remote containers only once
            if (!remoteContainerInvocation) {
                for (final String containerToUpdate : containersToUpdate) {
                    EndpointServiceImpl.this.log.debug("Updating remote new-endpoints (container: " + containerToUpdate + ") ...");

                    final List<Future<Long>> futures = new ArrayList<Future<Long>>(containersToUpdate.size()); 
                    // Submit the task updating new-endpoints on the remote container
                    final Callable<Long> updateRemoteNewEndpoints = new Callable<Long>() {
                        public Long call() {
                            try {
                                final DistributedJMXServer server = EndpointServiceImpl.this.jmxService
                                        .createDistributedJMXServer(containerToUpdate);

                                server.invoke(server.getEndpointServiceMBeanName(),
                                        DistributedJMXServer.UPDATE_NEW_ENDPOINTS, new Object[] { true },
                                        new String[] { boolean.class.getName() });

                                server.closeConnector();

                            } catch (final Exception e) {
                                EndpointServiceImpl.this.log.error("Failed to update new endpoints on the container '"
                                        + containerToUpdate + "'", e);
                            }

                            EndpointServiceImpl.this.log.debug("Remote new-endpoints updated (container: " + containerToUpdate + ").");
                            return null;
                        }
                    };
                    futures.add(ecs.submit(updateRemoteNewEndpoints));
                    remoteTaskNumber++;
                    this.log.debug("Updating remote new-endpoints (container: " + containerToUpdate + ") task added to the pool.");
                }
            }
            
            // Wait processing of local new-endpoints if needed
            if (newLocalEndpoints.size() > 0) {
                try {
                    updatedLocalNewEndpoints = updateLocalNewEndointsFuture.get();
                } catch (final InterruptedException e) {
                    this.log.warning("Local new-endpoints update has been interrupted.");
                } catch (final ExecutionException e) {
                    this.log.warning("An error occurs during the local new-endpoints update", e.getCause());
                }
            }
        }
            
        // Waits processing of remote new-endpoints update tasks
        for (int i = 0; i < remoteTaskNumber; i++) {
            try {
                ecs.take().get();
                
            } catch (final InterruptedException e) {
                this.log.warning("New endpoints update has been interrupted.");
            } catch (final ExecutionException e) {
                this.log.warning("An error occurs during the new endpoints update", e.getCause());
            }
        }

        this.log.end();
        return updatedLocalNewEndpoints;
    }

    protected boolean isNewEndpoints(ServiceEndpoint targetEndpoint) throws JBIException {
        this.log.call();

        return this.registry.isNewEndpoint(targetEndpoint);
    }

    /**
     * This method will update the given endpoint description and its
     * interfaces. The description DOM implementation provided by the component
     * endpoint is converted into the petals-kernel one to be able to use it
     * without a ClassNotFoundException.
     * 
     * @param endpoint
     *            endpoint to update
     */
    private void updateEndpointDescriptionAndInterfaces(final AbstractEndpoint endpoint) {
        this.log.start();

        Document internalEndpointDescription = null;
        List<QName> interfaces = new ArrayList<QName>();

        try {
            final Document componentEndpointDescription = this.adminService.getServiceDescription(endpoint.getComponentName(),
                    endpoint);
            
            // The description returns by the endpoint component uses the DOM
            // implementation of the endpoint component. We must convert it into
            // the JVM DOM implementation to be able to re-read it, next time,
            // in a component external context.
            internalEndpointDescription = DocumentBuilders
                    .getJvmDocumentBuilder().newDocument();
            
            final Transformer transformer = Transformers
					.getXmlWithDeclarationTransformer();
			try {
				transformer
				        .transform(new DOMSource(componentEndpointDescription),
				                new DOMResult(internalEndpointDescription));
			} finally {
				transformer.reset();
			}

        } catch (final PetalsException e) {
            QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Failed to retrieve the description from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName() + "'" + "'. Set its interface to " + interfaze, e);
            interfaces.add(interfaze);
        } catch (final TransformerException e) {
            QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Failed to convert the description of the endpoint '"
                    + endpoint.getEndpointName() + "' from the component '"
                    + endpoint.getComponentName() + "' into the internal DOM implementation. Set its interface to " + interfaze, e);
            interfaces.add(interfaze);
        }

        try {
            // Get the whole description
            if (interfaces.size() == 0) {
                if (internalEndpointDescription == null) {
                    throw new PetalsException("The description retrieved for the endpoint '"
                            + endpoint.getEndpointName() + "' is null");
                } else {
                    endpoint.setDescription(internalEndpointDescription);
                    interfaces = WSDLUtil.getInterfacesForService(internalEndpointDescription, endpoint
                            .getServiceName());
                }
            }
        } catch (PetalsException e) {
            QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Failed to retrieve interfaces from the description of the endpoint '"
                    + endpoint.getEndpointName() + "'. Set its interface to " + interfaze);
            interfaces.add(interfaze);
        }

        if (endpoint.getInterfaces() != null) {
            final List<QName> jbiInterfaces = Arrays.asList(endpoint.getInterfaces());
            if (!interfaces.containsAll(jbiInterfaces)) {
                final QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                        INCOHERENT_INTERFACE
						+ endpoint.getServiceName().getLocalPart());
                this.log.warning("Incoherent interface name for the endpoint "
                        + endpoint.getServiceName() + "-" + endpoint.getEndpointName());
                this.log.warning("\t- JBI Descriptor interface name: " +  endpoint.getInterfaces());
                this.log.warning("\t- WSDL interface name: " +  interfaces);
                interfaces.add(interfaze);
            }
        }
        else {
            // TODO: Is it possible that the interface is null ??
            endpoint.setInterfaces(interfaces.toArray(new QName[0]));
        }
        
        this.log.end();
    }

    private List<String> interfaceQNameToString(AbstractEndpoint abstractEndpoint) {
        QName[] interfaces = abstractEndpoint.getInterfaces();
        List<String> interfaceNames = new ArrayList<String>();
        if (interfaces != null && interfaces.length > 0) {
            for (QName interfaceName : interfaces) {
                interfaceNames.add(interfaceName.toString());
            }
        } else {
            interfaceNames.add(UNDEFINED);
        }
        return interfaceNames;
    }
    
    /**
     * Return all the registered endpoints for the specified container
     * 
     * @param containerName
     * @return a collection of AbstractEndpoint. Not null, may be empty
     * @throws NamingException
     *      problem accessing JNDI
     */
    public Collection<AbstractEndpoint> getServiceEndpointForContainer(final String containerName)
            throws NamingException {

        this.log.start();
        if (containerName == null) {
            throw new IllegalArgumentException("Container name must be non null");
        }
        if (0 == containerName.length()) {
            throw new IllegalArgumentException("Container name must be non empty");
        }

        final Collection<AbstractEndpoint> eps = this.registry.getServiceEndpointForContainer(containerName);

        this.log.end();
        
        return eps;
    }

}
