/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramRegistry.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.registry;

import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.naming.Binding;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.InvalidNameException;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.xml.namespace.QName;

import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.ExternalEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.InternalEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.LinkedEndpoint;
import org.objectweb.petals.util.JNDIUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.NamingHelper;
import org.objectweb.petals.util.SystemUtil;
import org.w3c.dom.Document;

/**
 * This registry is the registry used in a distributed environment. It used a
 * JNDI distributed directory to manage the endpoint storage. <br>
 * Only subdomain-endpoints or public endpoints from other subdomains are
 * retrieved by the getXXX() methods
 * 
 * This is the JNDIRegistry service format :
 * 
 * <pre>
 * '/'
 *  '/endpoints/'
 *  '          /e1/e1' (key is serviceName@endpointName)
 *  '              wsdl' (endpoint WSDL as DOM Document)
 *  '          /e2/e2'
 *  '              wsdl' (endpoint WSDL as DOM Document)'
 *  '/services/'
 *  '         /s1/ref-e1' (key is serviceName and object is link to /endpoints/e1)
 *  '         /s2/ref-e2'
 *  '/interfaces/'
 *  '           /i1/ref-e1' (key is itfName and object is link to /endpoints/e1)
 *  '           /i2/ref-e2'
 *  '/new-endpoints/'
 *  '              /ref-e1' (link to /endpoints/e1)
 *  '              /ref-e2' (link to /endpoints/e2)
 * </pre>
 * 
 * @author alouis - EBM Websourcing
 * @author ddesjardins - EBM Websourcing
 * @author chamerling - EBM Websourcing
 * @author rnaudin - EBM WebSourcing
 * @since Petals 1.1
 */
public class JNDIRegistry {

    /**
     * Name of the JNDI endpoints context
     */
    public static final String ENDPOINTS_REF = "endpoints";

    /**
     * Name of the JNDI endpoint WSDL object
     */
    public static final String ENDPOINT_WSDL_REF = "wsdl";

    /**
     * Name of the JNDI interfaces context
     */
    public static final String INTERFACES_REF = "interfaces";

    /**
     * Name of the JNDI new endpoints context
     */
    public static final String NEW_ENDPOINTS_REF = "new-endpoints";

    /**
     * Name of the JNDI service context
     */
    public static final String SERVICES_REF = "services";

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Endpoint context
     */
    protected Context endpointsContext;

    /**
     * Interfaces context
     */
    protected Context interfacesContext;

    /**
     * New endpoint context
     */
    protected Context newEndpointContext;

    /**
     * Root context
     */
    protected InitialContext rootContext;

    /**
     * Service context
     */
    protected Context servicesContext;

    /**
     * JNDI service fractal component
     */
    private JNDIService jndiService;

    /**
     * Endpoint service fractal component
     */
    private EndpointService endpointService;

    public JNDIRegistry() {
        super();
    }

    public JNDIRegistry(EndpointService endpointService, JNDIService jndiService, LoggingUtil log) {
        this();

        this.log = log;
        this.jndiService = jndiService;
        this.endpointService = endpointService;
    }

    /**
     * Unregister a connection
     * 
     * @param connection
     *      connection to unregistered; must be non null
     * @throws RegistryException
     *      if the connection was not previously registered
     */
    protected void unregisterConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException("The connection must be non null");
        }

        String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint.getToEndpointName());

        String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        if (linkedEndpoint.getInterfaces() != null) {
            Name interf = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);
            try {
                try {
                    // check if the provider endpoint is still bound in the
                    // new endpoint context if so its interfaces have not
                    // been resolved
                    this.newEndpointContext.lookup(toKey);
                    Context interfaceCtx = (Context) this.interfacesContext.lookup(interf);
                    interfaceCtx.unbind(key);
                } catch (NameNotFoundException e) {
                    // we need to check that we do not remove it from
                    // one of its interfaces context
                    InternalEndpoint internalEndpoint = (InternalEndpoint) this
                            .getInternalEndpoint(linkedEndpoint.getToServiceName(), linkedEndpoint
                                    .getToEndpointName());
                    if (internalEndpoint == null) {
                        throw new NameNotFoundException(linkedEndpoint.getEndpointName()
                                + " not found");
                    }
                    boolean isNotInInt = true;
                    for (QName i : internalEndpoint.getInterfaces()) {
                        if (linkedEndpoint.getInterfaces()[0].equals(i)) {
                            isNotInInt = false;
                        }
                    }
                    if (isNotInInt) {
                        Context interfaceCtx = (Context) this.interfacesContext.lookup(interf);
                        interfaceCtx.unbind(key);
                        if (!this.interfacesContext.list(interf).hasMoreElements()) {
                            this.interfacesContext.destroySubcontext(interf);
                        }
                    }
                }
            } catch (NamingException e) {
                throw new RegistryException("Failed to unregister connection with interface name "
                        + linkedEndpoint.getInterfaces()[0] + " to "
                        + linkedEndpoint.getToEndpointName(), e);
            }
        } else {
            try {
                // If the provider service name is different from the consumer
                // service name we removed it
                if (!linkedEndpoint.getToServiceName().equals(linkedEndpoint.getServiceName())) {
                    this.retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint.getServiceName())).unbind(key);
                    // If the service context is empty, we need to remove it
                    Name serviceName = NamingHelper.qNameToName(linkedEndpoint.getServiceName());
                    if (!this.servicesContext.list(serviceName).hasMore()) {
                        this.servicesContext.destroySubcontext(serviceName);
                    }
                }

                // If the provider endpoint name is different from the consumer
                // endpoint
                // name we removed it
                if (!toKey.equals(key)) {
                    Context endpointContext = (Context) this.endpointsContext.lookup(key);
                    endpointContext.unbind(key);
                    endpointContext.unbind(ENDPOINT_WSDL_REF);
                }

            } catch (NamingException e) {
                throw new RegistryException("Failed to unregister connection with service name "
                        + linkedEndpoint.getServiceName() + " and endpoint name "
                        + linkedEndpoint.getEndpointName() + " to "
                        + linkedEndpoint.getToEndpointName(), e);
            }
        }
        this.log.end();
    }

    /**
     * Unregister the given external endpoint.
     * 
     * @param externalEndpoint
     *      the external endpoint to be deregistered; must be non-null
     * @throws RegistryException
     *      if the endpoint was not previously registered
     */
    protected void unregisterExternalEndpoint(ServiceEndpoint externalEndpoint)
            throws RegistryException {
        this.log.call();

        this.unregisterInternalEndpoint((AbstractEndpoint) externalEndpoint);

    }

    /**
     * @see Registry#unregisterInternalEndpoint(AbstractEndpoint) <ol>
     *  <li>Unbind the endpoint from the endpoint context</li>
     *  <li>If the endpoint is still in the new-endpoints context, unbind the
     *  endpoint from the new-endpoints context</li>
     *  <li>Unbind the endpoint from the service context with contains it</li>
     *  <li>If the service subcontext is empty, remove it from the services
     *  context</li>
     *  <li>Unbind the endpoint from the interface context with contains it</li>
     *  <li>If the interface subcontext is empty, remove it from the interfaces
     *  context</li>
     *  </ol>
     */
    protected void unregisterInternalEndpoint(AbstractEndpoint internalEndpoint)
            throws RegistryException {
        this.log.start();

        if (internalEndpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }

        String key = getKey(internalEndpoint);

        try {

            // work either with the retrieved endpoint to be sure that the
            // getters of interfaces/description are functioning
            AbstractEndpoint retrievedEndpoint;

            // Verify that the endpoint is registered
            try {
                retrievedEndpoint = this.retrieveEndpoint(internalEndpoint);
            } catch (RegistryException e) {
                throw new RegistryException("The endpoint " + internalEndpoint.getEndpointName()
                        + " must be previously registered to be unregistered");
            }

            if (retrievedEndpoint == null) {
                throw new RegistryException("The endpoint " + internalEndpoint.getEndpointName()
                        + " can not be a link to be unregistered");
            }

            // Unbind either the newEndpoint or the registered interfaces
            try {
                this.newEndpointContext.lookup(key);
                this.newEndpointContext.unbind(key);
            } catch (NameNotFoundException e) {
                for (QName interfaceName : retrievedEndpoint.getInterfaces()) {
                    Name itfName = NamingHelper.qNameToName(interfaceName);
                    try {
                        Context interfaceCtx = (Context) this.interfacesContext.lookup(itfName);
                        interfaceCtx.unbind(key);

                        // delete subcontext if no more endpoint is found
                        if (!this.interfacesContext.list(itfName).hasMore()) {
                            this.interfacesContext.destroySubcontext(itfName);
                        }
                    } catch (NamingException e1) {
                        throw new RegistryException("Failed to unregister internal endpoint "
                                + retrievedEndpoint.getEndpointName() + ", its interface "
                                + interfaceName + " is not registered", e1);
                    }
                }
            }

            // Unbind from services context
            try {
                Name serviceName = NamingHelper.qNameToName(retrievedEndpoint.getServiceName());
                Context serviceCtx = (Context) this.servicesContext.lookup(serviceName);
                serviceCtx.unbind(key);

                // destroy service subcontext if no more endpoints serve
                // this services
                if (!this.servicesContext.list(serviceName).hasMore()) {
                    this.servicesContext.destroySubcontext(serviceName);
                }
            } catch (NamingException e1) {
                throw new RegistryException("Failed to unregister internal endpoint "
                        + retrievedEndpoint.getEndpointName() + ", its service "
                        + retrievedEndpoint.getServiceName() + " is not registered", e1);
            }

            // Finally unbind the endpoint
            Context endpointContext = (Context) this.endpointsContext.lookup(key);
            endpointContext.unbind(key);
            endpointContext.unbind(ENDPOINT_WSDL_REF);
            this.endpointsContext.destroySubcontext(key);

        } catch (NamingException e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * Get the external service endpoint for the named activated endpoint, if
     * any
     * 
     * @param service
     *      qualified name of the endpoint's service; must be non null
     * @param name
     *      name of the endpoint; must be non null
     * @return the named external endpoint, or null if the named endpoint is not
     *  activated
     */
    protected AbstractEndpoint getExternalEndpoint(QName serviceName, String endpointName)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(serviceName, endpointName,
                ExternalEndpoint.class.getName());

        return result;
    }

    /**
     * Queries for external endpoints that implements the given interface
     * 
     * @param interfaceName
     *      qualified name of the interface that contains the endpoints
     * @return Array of external endpoints
     * 
     *  Array of external endpoints that implements the qualified interface, if
     *  the interface name is non null
     * 
     *  Array of all the external endpoints, if the interface name is null
     * 
     * @throws RegistryException
     *      if a problem occurs
     */
    protected AbstractEndpoint[] getExternalEndpointsForInterface(QName interfaceName)
            throws RegistryException {
        this.log.call();
        return this.retrieveEndpointsForInterface(interfaceName, ExternalEndpoint.class.getName());
    }

    /**
     * Queries for external endpoints that are part of the given service
     * 
     * @param serviceName
     *      qualified name of the service that contains the endpoints; must be
     *      non null
     * 
     * @return array of available external endpoints for the specified service
     *  name; must be non null; may be empty
     * 
     */
    protected AbstractEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.call();
        return this.retrieveEndpointsForService(serviceName, ExternalEndpoint.class.getName());
    }

    /**
     * Get the internal service endpoint for the named activated endpoint, if
     * any. if the endpoint is a link, resolve it and return the real endpoint
     * 
     * @param service
     *      qualified name of the endpoint's service; must be non null
     * @param name
     *      name of the endpoint; must be non null
     * @return the named internal endpoint, or null if the named endpoint is not
     *  activated
     */
    protected AbstractEndpoint getInternalEndpoint(QName serviceName, String endpointName)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(serviceName, endpointName,
                InternalEndpoint.class.getName());

        return result;
    }

    /**
     * Get the internal service endpoint for the named activated endpoint, if
     * any
     * 
     * @param service
     *      qualified name of the endpoint's service; must be non null
     * @param name
     *      name of the endpoint; must be non null
     * @param resolveLink
     *      if the Endpoint is a link, resolve it (true) or return the link
     *      (false)
     * @return the named internal endpoint, or null if the named endpoint is not
     *  activated
     */
    protected AbstractEndpoint getInternalEndpoint(QName service, String name, boolean resolveLink)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(service, name, InternalEndpoint.class
                .getName(), resolveLink);

        return result;
    }

    /**
     * Queries for internal endpoints that implements the given interface
     * 
     * @param interfaceName
     *      qualified name of the interface that contains the endpoints
     * @return Array of internal endpoints
     * 
     *  Array of internal endpoints that implements the qualified interface, if
     *  the interface name is non null
     * 
     *  Array of all the external endpoints, if the interface name is null
     * 
     * @throws RegistryException
     *      if a problem occurs
     */
    protected AbstractEndpoint[] getInternalEndpointsForInterface(QName interfaceName)
            throws RegistryException {
        this.log.call();
        return this.retrieveEndpointsForInterface(interfaceName, InternalEndpoint.class.getName());
    }

    /**
     * Queries for internal endpoints that are part of the given service
     * 
     * @param serviceName
     *      qualified name of the service that contains the endpoints; must be
     *      non null
     * 
     * @return array of available internal endpoints for the specified service
     *  name; must be non null; may be empty
     * 
     */
    protected AbstractEndpoint[] getInternalEndpointsForService(QName serviceName) {
        this.log.call();
        return this.retrieveEndpointsForService(serviceName, InternalEndpoint.class.getName());
    }

    /**
     * Register a connection
     * 
     * @param connection
     *      connection to registered; must be non null
     * @throws RegistryException
     *      if the connection has already been registered or if a problem occurs
     */
    protected void registerConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException("The connection must be non null");
        }

        // target endpoint
        String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint.getToEndpointName());

        //' virtual' endpoint
        String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        try {
            LinkRef linkRef = new LinkRef(new CompositeName("").add(ENDPOINTS_REF).add(toKey).add(toKey));

            if (linkedEndpoint.getInterfaces() != null) {
                try {
                    Name itfName = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);

                    // bind or rebind link in interfaces
                    Context itfContext = this.retrieveInterfaceContext(itfName);
                    itfContext.rebind(key, linkRef);

                } catch (NamingException e) {
                    throw new RegistryException(
                            "Problem while registering connection with interfacename "
                                    + linkedEndpoint.getInterfaces()[0] + " to "
                                    + linkedEndpoint.getToEndpointName() + " ");
                }
            } else {
                try {
                    Name serviceName = NamingHelper.qNameToName(linkedEndpoint.getServiceName());

                    // bind or rebind service
                    Context servContext = this.retrieveServiceContext(serviceName);
                    servContext.rebind(key, linkRef);

                    // bind or rebind endpoint
                    this.endpointsContext.rebind(key, linkRef);

                } catch (NamingException e) {
                    throw new RegistryException(
                            "Problem while registering connection with service name  "
                                    + linkedEndpoint.getServiceName() + " and endpoint name "
                                    + linkedEndpoint.getEndpointName() + " to "
                                    + linkedEndpoint.getToEndpointName() + " ");
                }
            }
        } catch (final InvalidNameException e) {
            throw new RegistryException(
                    "Problem while registering connection  "
                            + linkedEndpoint.getServiceName() + " and endpoint name "
                            + linkedEndpoint.getEndpointName() + " to "
                            + linkedEndpoint.getToEndpointName() + " ");
        } finally {
            this.log.end();
        }
    }

    /**
     * Register the given external endpoint
     * 
     * @param externalEndpoint
     *      the external endpoint to be registered; must be non null
     * @throws RegistryException
     *      if an external endpoint with the same name is already registered
     */
    protected void registerExternalEndpoint(ServiceEndpoint externalEndpoint)
            throws RegistryException {
        this.registerEndpoint(externalEndpoint);
    }

    /**
     * Register the given internal endpoint
     * 
     * @param internalEndpoint
     *      the internal endpoint to be registered; must be non null
     * @throws RegistryException
     *      if an internal endpoint with the same name is already registered
     */
    protected void registerInternalEndpoint(AbstractEndpoint endpoint) throws RegistryException {
        this.registerEndpoint(endpoint);
    }

    /**
     * Retrieve the list of all the new endpoints. This method is synchronized
     * to avoid concurrent invocations.
     * 
     * @param localNewEndpoints
     *      The list of the local new endpoints to fill
     * @param containerToUpdate
     *      The list of containerToUpdate to fill
     * 
     * @throws RegistryException
     */
    protected synchronized void retrieveNewEndpoints(List<AbstractEndpoint> localNewEndpoints,
            List<String> containerToUpdate) throws RegistryException {
        this.log.start();

        try {
            NamingEnumeration<Binding> bindings = this.rootContext.listBindings(NEW_ENDPOINTS_REF);

            while (bindings.hasMoreElements()) {
                Binding binding = bindings.next();
                Object object = binding.getObject();

                if (object instanceof LinkRef) {
                    final CompositeName newEndpointName = new CompositeName(((LinkRef) object).getLinkName());
                    
                    try {
                        AbstractEndpoint abstractEndpoint = (AbstractEndpoint) this.rootContext
                                .lookup(newEndpointName);
                        if (abstractEndpoint.getContainerName().equals(
                                SystemUtil.getContainerName())) {
                            localNewEndpoints.add(abstractEndpoint);
                        } else if (!containerToUpdate.contains(abstractEndpoint.getContainerName())) {
                            containerToUpdate.add(abstractEndpoint.getContainerName());
                        }

                    } catch (NameNotFoundException e1) {
                        // The new endpoint may be updated concurrently by
                        // another node, do nothing
                        this.log.error("The current endpoint [" + newEndpointName.toString() + "] may be updated concurrently (" + e1.getMessage() + ")");
                    }
                } else {
                    this.log.error("The new endpoint entry named '" + binding.getName()
                            + "' must be a " + LinkRef.class + " instance. It will be deleted");
                    this.newEndpointContext.unbind(binding.getName());
                }
            }

        } catch (NamingException e) {
            throw new RegistryException("Problem while retrieving new endpoints", e);
        }

        this.log.end();
    }

    /**
     * Update the endpoint, remove it from the new-endpoint context and update
     * the interfaces context
     * 
     * @param endpoint
     *      endpoint to validate; must be non null
     * @throws RegistryException
     *      if a problem occurs
     */
    protected void validateEndpoint(final AbstractEndpoint endpoint) throws RegistryException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }

        final String key = JNDIRegistry.getKey(endpoint);

        try {
            // Add the interfaces to the interfaces context
            for (final QName name : endpoint.getInterfaces()) {
                Name itfName = NamingHelper.qNameToName(name);
                final Context interfaceCtx = this.retrieveInterfaceContext(itfName);

                // bind or rebind endpoint in interfaces
                final LinkRef linkRef = new LinkRef(new CompositeName("").add(ENDPOINTS_REF).add(key).add(key));
                interfaceCtx.rebind(key, linkRef);
            }

            // Update the endpoint in the endpoint context
            Context endpointContext = (Context) this.endpointsContext.lookup(key);
            endpointContext.rebind(key, endpoint);
            endpointContext.rebind(ENDPOINT_WSDL_REF, endpoint.getDescription());

        } catch (NamingException e) {
            throw new RegistryException(e);
        } finally {
            try {
                // We remove the endpoint under the context "new-endpoints"
                this.newEndpointContext.unbind(key);
            } catch (final NameNotFoundException e) {
                this.log.warning("The endpoint [" + key + "] does not exist anymore under the context " + NEW_ENDPOINTS_REF);
                // do nothing
            } catch (final NamingException e) {
                this.log.error(
                        "An error occurs during the unbind of the endpoint ["
                                + key + "] under the context "
                                + NEW_ENDPOINTS_REF + ": " + e.getMessage(), e);
                // do nothing
            }
        }

        this.log.end();
    }

    /**
     * Create and / or reference contexts "containers", "endpoints", "services"
     * and "interfaces" in the global directory.
     * 
     * @throws JBIException
     *      problems while creating or lookup contexts
     */
    protected void initContexts() throws RegistryException {
        this.log.start();
        try {
            this.rootContext = this.jndiService.getInitialContext();

            try {
                this.endpointsContext = (Context) this.rootContext.lookup(ENDPOINTS_REF);
            } catch (NameNotFoundException e) {
            	// Caution: In a PEtALS network, a distributed JNDI
				// synchronization can arrive between the check of the endpoint
				// context registration and its registration. This throws a
				// NameAlreadyBoundException that must be catched correctly.
                try {
                	this.endpointsContext = this.rootContext.createSubcontext(ENDPOINTS_REF);
                }
                catch (final NameAlreadyBoundException e1) {
                	this.endpointsContext = (Context) this.rootContext.lookup(ENDPOINTS_REF);
                }
            }

            try {
                this.servicesContext = (Context) this.rootContext.lookup(SERVICES_REF);
            } catch (NameNotFoundException e) {
            	// Caution: In a PEtALS network, a distributed JNDI
				// synchronization can arrive between the check of the services
				// context registration and its registration. This throws a
				// NameAlreadyBoundException that must be catched correctly.
                try {
                	this.servicesContext = this.rootContext.createSubcontext(SERVICES_REF);
                }
                catch (final NameAlreadyBoundException e1) {
                	this.servicesContext = (Context) this.rootContext.lookup(SERVICES_REF);
                }
            }

            try {
                this.newEndpointContext = (Context) this.rootContext.lookup(NEW_ENDPOINTS_REF);
            } catch (NameNotFoundException e) {
            	// Caution: In a PEtALS network, a distributed JNDI
				// synchronization can arrive between the check of the new
				// endpoints context registration and its registration. This
				// throws a NameAlreadyBoundException that must be catched
				// correctly.
                try {
                	this.newEndpointContext = this.rootContext.createSubcontext(NEW_ENDPOINTS_REF);
                }
                catch (final NameAlreadyBoundException e1) {
                	this.newEndpointContext = (Context) this.rootContext.lookup(NEW_ENDPOINTS_REF);
                }
            }

            try {
                this.interfacesContext = (Context) this.rootContext.lookup(INTERFACES_REF);
            } catch (NameNotFoundException e) {
            	// Caution: In a PEtALS network, a distributed JNDI
				// synchronization can arrive between the check of the interfaces
				// context registration and its registration. This throws a
				// NameAlreadyBoundException that must be catched correctly.
                try {
                	this.interfacesContext = this.rootContext.createSubcontext(INTERFACES_REF);
                }
                catch (final NameAlreadyBoundException e1) {
                	this.interfacesContext = (Context) this.rootContext.lookup(INTERFACES_REF);
                }
            }

            if (this.endpointsContext == null || this.servicesContext == null
                    || this.interfacesContext == null) {
                throw new NotContextException();
            }
        } catch (Exception e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * Register a new Endpoint.
     * 
     * @param endpoint
     * @throws RegistryException
     */
    protected void registerEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must not be null");
        }

        AbstractEndpoint registeredEndpoint = null;
        boolean found = false;

        try {
            // Retrieve a potential matching endpoint
            registeredEndpoint = this.retrieveEndpoint(endpoint);
            found = true;
        } catch (RegistryException e) {
            String key = getKey(endpoint);

            try {
                // We register the endpoint
                
                // Caution: In a PEtALS network, a distributed JNDI
                // synchronization can arrive between the check of the endpoint
                // registration and its registration. This throws a
                // NameAlreadyBoundException that must be catched correctly.
                Context endpointContext = this.endpointsContext.createSubcontext(key);
                endpointContext.bind(key, endpoint);

                LinkRef linkRef = new LinkRef(new CompositeName("").add(ENDPOINTS_REF).add(key).add(key));

                // register service
                Name serviceName = NamingHelper.qNameToName(endpoint.getServiceName());
                if (serviceName != null) {
                    Context servContext = this.retrieveServiceContext(serviceName);
                    try {
                        servContext.bind(key, linkRef);
                    } catch (NameAlreadyBoundException e2) {
                        // do nothing
                    }
                }

                // register the newEndpoint only at the end
                this.newEndpointContext.bind(key, linkRef);

            } catch (final NameAlreadyBoundException e1) {
                // See previous comment on distributed JNDI synchronization
                registeredEndpoint = this.retrieveEndpoint(endpoint);
                found = true;
                
            } catch (final NamingException e1) {
                throw new RegistryException("Problem while registering endpoint "
                        + endpoint.getEndpointName(), e1);
            }
        }

        if (found) {
            // registeredEndpoint == null ==> the bound object isn't a real
            // endpoint (link or anything else)
            if (registeredEndpoint == null) {
                throw new RegistryException(
                        "You are trying to register an endpoint with a name that "
                                + "is already bound to an other object "
                                + "(link or anything else) in the JNDI Registry.");
            }

            // An endpoint can't be registered with the same endpoint name as a
            // previously registered endpoint that have different container name
            // or component name
            if (!(((AbstractEndpoint) endpoint).getContainerName().equals(
                    registeredEndpoint.getContainerName()) && ((AbstractEndpoint) endpoint)
                    .getComponentName().equals(registeredEndpoint.getComponentName()))) {
                throw new RegistryException("The endpoint " + endpoint.getEndpointName()
                        + " is already registered for an other component/container");
            }
        }

        this.log.end();
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any. If the Endpoint is a link, resolve it and return the
     * real endpoint
     * 
     * @param serviceName
     *      the {@link QName} of the service
     * @param name
     *      the endpoint name
     * @param endpointClassName
     *      class name of the endpoint
     * @return the matching endpoint
     * @throws RegistryException
     */
    private AbstractEndpoint retrieveEndpoint(QName serviceName, String endpointName,
            String endpointClassName) throws RegistryException {
        this.log.call();
        AbstractEndpoint endpoint = this.retrieveEndpoint(serviceName, endpointName,
                endpointClassName, true);
        return endpoint;
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any
     * 
     * @param service
     *      qualified service name
     * @param endpointName
     *      endpoint name
     * @param endpointClassName
     *      class name of the endpoint
     * @param resolveLink
     *      if the Endpoint is a link, resolve it (true) or return the link
     *      (false)
     * @return
     * @throws RegistryException
     */
    private AbstractEndpoint retrieveEndpoint(QName serviceName, String endpointName,
            String endpointClassName, boolean resolveLink) throws RegistryException {
        this.log.start();

        AbstractEndpoint result = null;
        String key = getKey(serviceName, endpointName);

        try {
            Context endpointContext = (Context) this.endpointsContext.lookup(key);
            Object outObj = endpointContext.lookup(key);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                AbstractEndpoint endpoint = (AbstractEndpoint) outObj;
                if (endpoint.getClass().getName().equals(endpointClassName)) {
                    endpoint.setEndpointService(this.endpointService);
                    
                    // retrieve endpoint description
                    final Object docObj = endpointContext.lookup(ENDPOINT_WSDL_REF);
                    if (docObj instanceof Document) {
                        endpoint.setDescription((Document)docObj);
                    }
                    
                    result = endpoint;
                }
            }
            // link endpoint
            else if (outObj instanceof LinkRef) {
                Object obj = this.rootContext.lookup(((LinkRef) outObj).getLinkName());
                if (obj instanceof AbstractEndpoint) {
                    AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                    if (endpoint.getClass().getName().equals(endpointClassName)) {
                        // return the real endpoint
                        if (resolveLink) {
                            endpoint.setEndpointService(this.endpointService);
                            result = endpoint;
                        }
                        // return a link
                        else {
                            result = new LinkedEndpoint(serviceName, endpointName, endpoint
                                    .getServiceName(), endpoint.getEndpointName(), null);
                        }
                    }
                }
            }
        } catch (NamingException e1) {
            throw new RegistryException("Error while looking up an Endpoint in the JNDI Registry",
                    e1);
        }

        this.log.end();
        return result;
    }

    /**
     * Retrieve an endpoint in the JNDI registry without making distinction
     * between Internal and External one.
     * 
     * @param endpoint
     *      endpoint description
     * @return an {@link AbstractEndpoint} if the matching endpoint is a real
     *  endpoint, null otherwise (for exemple, the matching endpoint is a link)
     * 
     * @throws RegistryException
     */
    private AbstractEndpoint retrieveEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();

        AbstractEndpoint result = null;
        try {
            String key = getKey(endpoint);
            Context endpointContext = (Context) this.endpointsContext.lookup(key);
            Object outObj = endpointContext.lookup(key);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                result = (AbstractEndpoint) outObj;
                result.setEndpointService(this.endpointService);
            }
        } catch (NamingException e) {
            throw new RegistryException("Error while looking up an Endpoint in the JNDI Registry",
                    e);
        }

        this.log.end();
        return result;
    }

    /**
     * @see Registry#getExternalEndpointsForService(QName)
     */
    private AbstractEndpoint[] retrieveEndpointsForService(QName serviceName, String endpointType) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("The service name must be non null");
        }

        AbstractEndpoint[] result = new AbstractEndpoint[0];
        Name srvName = NamingHelper.qNameToName(serviceName);
        try {
            NamingEnumeration<Binding> bindings = this.servicesContext.listBindings(srvName);
            List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
            while (bindings.hasMoreElements()) {
                Binding binding = bindings.next();
                Object object = binding.getObject();
                if (object instanceof LinkRef) {
                    LinkRef linkRef = (LinkRef) object;
                    Object obj = this.rootContext.lookup(linkRef.getLinkName());
                    if (obj instanceof AbstractEndpoint) {
                        AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                        if (endpoint.getClass().getName().equals(endpointType)) {
                            endpoint.setEndpointService(this.endpointService);
                            endpoints.add(endpoint);
                        }
                    }
                }
            }
            result = endpoints.toArray(new AbstractEndpoint[0]);
        } catch (NamingException e) {
            this.log.error(e);
        }

        this.log.end();
        return result;
    }

    protected boolean isNewEndpoint(ServiceEndpoint serviceEndpoint) {
        this.log.call();

        try {
            this.newEndpointContext.lookup(getKey(serviceEndpoint));
            return true;
        } catch (NamingException e) {
            return false;
        }
    }

    protected void cleanNewEndpoints(List<AbstractEndpoint> newEndpoints) {
        this.log.start();

        for (AbstractEndpoint newEndpoint : newEndpoints) {
            try {
                this.newEndpointContext.unbind(getKey(newEndpoint));
            } catch (NamingException e) {
                this.log.warning("Failed to unbind the new endpoint '"
                        + newEndpoint.getEndpointName() + "'");
            }
        }

        this.log.end();
    }

    /**
     * Get a unique key for the service endpoint. As the character
     * <code>'/'</code> is usually used as separator in the JNDI composite name,
     * for confusing reasons in the endpoint name, it is needed to replace this
     * character (<code>'/'</code>) in endpoint name.
     * 
     * @param ep
     * @return
     */
    private static final String getKey(final ServiceEndpoint ep) {
        return JNDIRegistry.getKey(ep.getServiceName(), ep.getEndpointName());
    }

    /**
     * Create an unique key for the serviceName / endpointName.
     * 
     * @see {@link JNDIRegistry#getKey(QName, String)}
     * @return
     */
    private static final String getKey(QName serviceName, String endpointName) {
        String result = serviceName + "@" + endpointName;
        return result.replaceAll("/", "&47;");
    }

    /**
     * Retrieve the service name from a jndi key
     * 
     * @see {@link JNDIRegistry#getKey(QName, String)}
     * @return
     */
    private static final QName getServiceNameFromKey(String key) {
        String[] result = key.split("@");
        return QName.valueOf(result[0].replaceAll("&47;", "/"));
    }

    /**
     * Retrieve the endpoint name from a jndi key
     * 
     * @see {@link JNDIRegistry#getKey(QName, String)}
     * 
     * @return
     */
    private static final String getEndpointNameFromKey(String key) {
        String[] result = key.split("@");
        return result[1].replaceAll("&47;", "/");
    }

    /**
     * Return the interfaceContext associated with this interfaceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    private Context retrieveInterfaceContext(Name interfaceName) throws NamingException {
        this.log.call();

        Context interfContext = null;
        try {
            interfContext = (Context) this.interfacesContext.lookup(interfaceName);
        } catch (NameNotFoundException e) {
            interfContext = this.interfacesContext.createSubcontext(interfaceName);
        }
        return interfContext;
    }

    /**
     * Return the serviceContext associated with this serviceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    private Context retrieveServiceContext(Name serviceName) throws NamingException {
        this.log.call();

        Context serviceContext = null;

        try {
            serviceContext = (Context) this.servicesContext.lookup(serviceName);
        } catch (NameNotFoundException e) {
            serviceContext = this.servicesContext.createSubcontext(serviceName);
        }

        return serviceContext;
    }

    /**
     * Retrieve the list of endpoint that implements the interface name
     * 
     * @param interfaceName
     *      the {@link QName} of the interface
     * @param endpointType
     *      the type (class name) of the endpoint (InternalEndpoint or
     *      ExternalEndpoint)
     * @return an Array of matching endpoints
     * @throws RegistryException
     */
    private AbstractEndpoint[] retrieveEndpointsForInterface(QName interfaceName,
            String endpointType) throws RegistryException {
        this.log.start();
        List<AbstractEndpoint> result = new ArrayList<AbstractEndpoint>();
        try {
            if (interfaceName == null) {
                // We must return all the endpoints of the specified type
                NamingEnumeration<Binding> bindingsContextEndpoint = this.rootContext.listBindings(ENDPOINTS_REF);

                while (bindingsContextEndpoint.hasMoreElements()) {
                    final Binding bindingContextEndpoint = bindingsContextEndpoint.next();
                    final Object objectContextEndpoint = bindingContextEndpoint.getObject();
                    
                    if (objectContextEndpoint instanceof Context) {
                        NamingEnumeration<Binding> bindingsEndpoint = this.endpointsContext.listBindings(bindingContextEndpoint.getName());
                        
                        while (bindingsEndpoint.hasMoreElements()) {
                            final Binding bindingEndpoint = bindingsEndpoint.next();
                            final Object objectEndpoint = bindingEndpoint.getObject();
                            
                            // real endpoint case
                            if (objectEndpoint instanceof AbstractEndpoint) {
                                AbstractEndpoint endpoint = (AbstractEndpoint) objectEndpoint;
                                if (endpoint.getClass().getName().equals(endpointType)) {
                                    endpoint.setEndpointService(this.endpointService);
                                    result.add(endpoint);
                                }
                            }
                            // connection endpoint case
                            else if (objectEndpoint instanceof LinkRef) {
                                LinkRef linkRef = (LinkRef) objectEndpoint;
                                Object obj = this.rootContext.lookup(linkRef.getLinkName());
                                // connection linked to a real EP
                                if (obj instanceof AbstractEndpoint) {
                                    AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                                    if (endpoint.getClass().getName().equals(endpointType)) {
                                        LinkedEndpoint epLink = new LinkedEndpoint(
                                                getServiceNameFromKey(bindingEndpoint.getName()),
                                                getEndpointNameFromKey(bindingEndpoint.getName()), endpoint
                                                        .getServiceName(), endpoint.getEndpointName(), null);
                                        result.add(epLink);
                                    }
                                }
                            }
                        }
                    }
                }
            } else {
                // We must return only the endpoints for the specified
                // interface
                Name itfName = NamingHelper.qNameToName(interfaceName);
                NamingEnumeration<Binding> bindings = this.interfacesContext.listBindings(itfName);

                while (bindings.hasMoreElements()) {
                    Binding binding = bindings.next();
                    Object object = binding.getObject();
                    if (object instanceof LinkRef) {
                        LinkRef linkRef = (LinkRef) object;
                        Object obj = this.rootContext.lookup(linkRef.getLinkName());
                        if (obj instanceof AbstractEndpoint) {
                            AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                            if (endpoint.getClass().getName().equals(endpointType)) {
                                endpoint.setEndpointService(this.endpointService);
                                result.add(endpoint);
                            }
                        }
                    }
                }
            }
        } catch (NamingException e) {
            throw new RegistryException(
                    "Problem while getting endpoints for the specified interface " + interfaceName,
                    e);
        }

        this.log.end();
        return result.toArray(new AbstractEndpoint[0]);
    }

    /**
     * Return all the registered endpoints for the specified container
     * 
     * @param containerName
     * @return a collection of AbstractEndpoint. Not null, may be empty
     * @throws NamingException
     *      problem accessing JNDI
     */
    protected Collection<AbstractEndpoint> getServiceEndpointForContainer(final String containerName)
            throws NamingException {

        this.log.start();

        final Collection<AbstractEndpoint> eps = new HashSet<AbstractEndpoint>();

        if (JNDIUtil.isBound(this.rootContext, "", ENDPOINTS_REF)) {
            
            // We must return all the endpoints
            NamingEnumeration<Binding> bindingsContextEndpoint = this.rootContext.listBindings(ENDPOINTS_REF);

            while (bindingsContextEndpoint.hasMoreElements()) {
                final Binding bindingContextEndpoint = bindingsContextEndpoint.next();
                final Object objectContextEndpoint = bindingContextEndpoint.getObject();
                
                if (objectContextEndpoint instanceof Context) {
                    NamingEnumeration<Binding> bindingsEndpoint = this.endpointsContext.listBindings(bindingContextEndpoint.getName());
                    
                    while (bindingsEndpoint.hasMoreElements()) {
                        final Binding bindingEndpoint = bindingsEndpoint.next();
                        final Object objectEndpoint = bindingEndpoint.getObject();
                        
                        // endpoint case
                        if (objectEndpoint instanceof AbstractEndpoint) {
                            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) objectEndpoint;
                            if (abstractEndpoint.getContainerName().equals(containerName)) {
                                eps.add(abstractEndpoint);
                            }
                        }
                    }
                }
            }
        }
        this.log.end();

        return eps;
    }
}
