/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RouterImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.routing;

import java.util.Collection;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.registry.EndpointService;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.ConsumerEndpoint;
import org.objectweb.petals.jbi.messaging.transport.TransportException;
import org.objectweb.petals.jbi.messaging.transport.Transporter;
import org.objectweb.petals.util.InputStreamForker;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Routes messages to their destinations. Performs an itinerary resolution,
 * adding for example a transformatino service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.jbi.messaging.routing.Router.class))
public class RouterImpl implements Router {

	/**
	 * the endpoint service.
	 */
	@Requires(name = "endpoint", signature = org.objectweb.petals.jbi.messaging.registry.EndpointService.class)
	private EndpointService endpointService;

	/**
	 * logger wrapper.
	 */
	private LoggingUtil log;

	/**
	 * the logger.
	 */
	@Monolog(name = "logger")
	private Logger logger;

	/**
	 * the transporter.
	 */
	@Requires(name = "transporter", signature = org.objectweb.petals.jbi.messaging.transport.Transporter.class)
	private Transporter transporter;

	/**
	 * the container network.
	 */
	@Requires(name = "network", signature = org.objectweb.petals.communication.network.NetworkService.class)
	private NetworkService containerNetwork;

	/**
	 * the address resolver.
	 */
	private AddressResolver addressResolver;

	private Map<String, Map<String, InputStreamForker>> exchangeForkedStreamCache;

	private final static boolean highAvailibility = Boolean
			.parseBoolean(SystemUtil.getHA());

	/**
	 * the default contructor.
	 * 
	 */
	public RouterImpl() {
	}

	// ----------------------------------------------------------------------
	// Fractal LifecycleController implementation
	// ----------------------------------------------------------------------

	/**
	 * start the router.
	 * 
	 * @throws RoutingException
	 * 		impossible to start the router
	 */
	@LifeCycle(on = LifeCycleType.START)
	protected void start() throws RoutingException {
		this.log = new LoggingUtil(this.logger);
		this.log.start();

		this.addressResolver = new AddressResolver(this.containerNetwork,
				this.endpointService, this.logger);

		this.exchangeForkedStreamCache = new ConcurrentHashMap<String, Map<String, InputStreamForker>>();

		this.log.end();
	}

	/**
	 * stop the router.
	 */
	@LifeCycle(on = LifeCycleType.STOP)
	protected void stop() {
		this.log.call();

		this.addressResolver = null;
	}

	/**
	 * Sends the specified message. Depending on the Role of the sender, send
	 * the message to the producer (request) or to the consumer (response)
	 * 
	 * @param source
	 * 		the source
	 * @param exchange
	 * 		the exchange
	 * @param timeOut
	 * 		timeout of the message (in ms). 0 for no timeout
	 * @throws RoutingException
	 * 		impossible to send the message
	 */
	public void send(final ComponentContextImpl source,
			final MessageExchangeImpl exchange, final long timeOut)
			throws RoutingException {
		this.log.call();

		if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
			this.sendToProvider(source, exchange, timeOut, false);
		} else {
			this.sendToConsumer(source, exchange, timeOut, false);
		}
	}

	/**
	 * Send the message synchronously into the NMR.
	 * 
	 * @param source
	 * 		the source
	 * @param exchange
	 * 		the exchange
	 * @param timeOut
	 * 		timeout of the message (in ms). 0 for no timeout
	 * 
	 * @return the reponse exchange if received, or <code>null</code> if the
	 * 	timeout is reached
	 * 
	 * @throws RoutingException
	 * 		impossible to realize a send synchrone
	 */
	public MessageExchangeImpl sendSync(final ComponentContextImpl source,
			final MessageExchangeImpl exchange, final long timeOut)
			throws RoutingException {
		this.log.call();

		MessageExchangeImpl responseExchange = null;

		if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
			responseExchange = this.sendToProvider(source, exchange, timeOut,
					true);
		} else {
			responseExchange = this.sendToConsumer(source, exchange, timeOut,
					true);
		}

		return responseExchange;
	}

	/**
	 * Sends the specified message, sent by the specified component. TODO send
	 * directly to the DeliveryChannel of the component related to the recipient
	 * endpoint.
	 * 
	 * @param source
	 * 		the source
	 * @param exchange
	 * 		the exchange
	 * @param timeOut
	 * 		timeout of the message (in ms). 0 for no timeout
	 * @param sync
	 * 		true if the message is synchrone
	 * @return the reponse exchange if a synchronious send is processed
	 * @throws RoutingException
	 * 		impossible to send the message to the consumer
	 */
	protected MessageExchangeImpl sendToConsumer(
			final ComponentContextImpl source,
			final MessageExchangeImpl exchange, final long timeout,
			final boolean sync) throws RoutingException {
		this.log.start();

		MessageExchangeImpl responseExchange = null;

		try {
			ConsumerEndpoint consumerEndpoint = exchange.getConsumerEndpoint();
			exchange.setRole(MessageExchange.Role.CONSUMER);

			if (sync) {
				responseExchange = this.transporter.sendSync(exchange,
						consumerEndpoint.getComponentName(), consumerEndpoint
								.getContainerName(), timeout);
				// The timeout is reached
				if (responseExchange == null) {
					exchange.setRole(MessageExchange.Role.PROVIDER);
				}
			} else {
				this.transporter.send(exchange, consumerEndpoint
						.getComponentName(), consumerEndpoint
						.getContainerName(), timeout);
			}
		} catch (TransportException e) {
			// Restore the Role if the message has not been sent
			exchange.setRole(MessageExchange.Role.PROVIDER);
			throw new RoutingException(e);
		}

		this.log.end();
		return responseExchange;
	}

	/**
	 * Sends the specified message to the provider of the service. TODO for the
	 * moment, send directly to the DeliveryChannel of the component related to
	 * the recipient endpoint.
	 * 
	 * @param source
	 * 		the source
	 * @param exchange
	 * 		the exchange
	 * @param timeOut
	 * 		timeout of the message (in ms). 0 for no timeout
	 * @param sync
	 * 		true if the message is synchrone
	 * @return the reponse exchange if a synchronious send is processed
	 * @throws RoutingException
	 * 		impossible to send the message to the provider
	 */
	protected MessageExchangeImpl sendToProvider(
			final ComponentContextImpl source,
			final MessageExchangeImpl exchange, final long timeOut,
			final boolean sync) throws RoutingException {
		this.log.start();

		MessageExchangeImpl responseExchange = null;
		Exception exception = null;

		// resolve the address (find potential endpoints)
		final Collection<WeightedEndpoint> weightedEndpoints = this.addressResolver
				.resolveAddress(source, exchange, sync);

		exchange.setRole(MessageExchange.Role.PROVIDER);

		short counter = 0;
		for (WeightedEndpoint weightedEndpoint : weightedEndpoints) {

			log.debug("Attempt to route the exchange to the endpoint '"
					+ weightedEndpoint.endpoint.getEndpointName() + "'");

			// initialize the send attempt
			counter++;
			exception = null;
			exchange.setEndpoint(weightedEndpoint.endpoint);

			// fork the source only if necessary
			if (highAvailibility && counter != weightedEndpoints.size()) {
				try {
					this.forkExchangeSources(exchange);
				} catch (final MessagingException e) {
					exception = e;
					break;
				}
			}

			try {
				if (sync) {
					// We don't wheck that the remote container is active.
					// Inactive containers are processed as exception.
					responseExchange = this.transporter.sendSync(exchange,
							((AbstractEndpoint) weightedEndpoint.endpoint)
									.getComponentName(),
							((AbstractEndpoint) weightedEndpoint.endpoint)
									.getContainerName(), timeOut);
					// the timeout is reached is the return exchange is null
					if (responseExchange == null) {
						exchange.setRole(MessageExchange.Role.CONSUMER);
					}
				} else {
					this.transporter.send(exchange,
							((AbstractEndpoint) weightedEndpoint.endpoint)
									.getComponentName(),
							((AbstractEndpoint) weightedEndpoint.endpoint)
									.getContainerName(), timeOut);
				}
				// transfer terminated
				break;
			} catch (final TransportException e) {
				exception = e;
				if (highAvailibility) {
					this.log.warning(e);
				} else {
					break;
				}
			}
		}

		// clear the cache for this exchange
		if (highAvailibility) {
			this.exchangeForkedStreamCache.remove(exchange.getExchangeId());
		}

		if (exception != null) {
			// the transfer attempts finished all by a failure
			exchange.setRole(MessageExchange.Role.CONSUMER);
			if (exception instanceof RoutingException) {
				throw (RoutingException) exception;
			} else {
				throw new RoutingException(exception);
			}

		}

		this.log.end();
		return responseExchange;
	}

	/**
	 * Receive a message from the Transporter and return it to the Delivery
	 * Channel.
	 * 
	 * @param source
	 * 		The source component context
	 * @param timeoutMS
	 * 		The timeout to receive the incoming message
	 * @param block
	 * 		If <code>True</code>, the receive is blocking until a response
	 * @throws RoutingException
	 * 		impossible to receive a message
	 * @return the received message
	 */
	public MessageExchangeImpl receive(final ComponentContextImpl source,
			final long timeoutMS) throws RoutingException {
		this.log.start();

		MessageExchangeImpl exchange = null;
		String componentName = source.getComponentName();

		try {
			exchange = this.transporter.receive(componentName, timeoutMS);
		} catch (TransportException e) {
			throw new RoutingException(e);
		}

		this.log.end();

		return exchange;
	}

	/**
	 * (non-Javadoc).
	 * 
	 * @see
	 * 	org.objectweb.petals.jbi.routing.Router#addComponent(java.lang.String)
	 * @param componentName
	 * 		the name of the component
	 * @throws RoutingException
	 * 		impossible to add the component
	 */
	public void addComponent(final String componentName)
			throws RoutingException {
		this.log.start();

		try {
			this.transporter.addDestination(componentName);
		} catch (TransportException e) {
			throw new RoutingException(e);
		}

		this.log.end();
	}

	/**
	 * (non-Javadoc).
	 * 
	 * @see
	 * 	org.objectweb.petals.jbi.routing.Router#removeComponent(java.lang.String
	 * 	)
	 * @param componentName
	 * 		the name of the component
	 * @throws RoutingException
	 * 		impossible to remove the component
	 */
	public void removeComponent(final String componentName)
			throws RoutingException {
		this.log.start();

		try {
			this.transporter.removeDestination(componentName);
		} catch (TransportException e) {
			throw new RoutingException(e);
		}

		this.log.end();
	}

	/**
	 * Get the logger
	 * 
	 * @return the logger
	 */
	protected Logger getLogger() {
		return this.logger;
	}

	/**
	 * Handle streaming issue.
	 * 
	 * @param exchange
	 * @throws MessagingException
	 */
	private void forkExchangeSources(MessageExchangeImpl exchange)
			throws MessagingException {

		Map<String, InputStreamForker> exchangeStreamForked = exchangeForkedStreamCache
				.get(exchange.getExchangeId());
		if (exchangeStreamForked == null) {
			exchangeStreamForked = new ConcurrentHashMap<String, InputStreamForker>();
			exchangeForkedStreamCache.put(exchange.getExchangeId(),
					exchangeStreamForked);
		}

		Map<String, NormalizedMessage> messages = exchange.getMessages();
		for (String messageName : messages.keySet()) {
			// TODO Handle the Streaming issue for the Attachments??
			Source content = messages.get(messageName).getContent();
			if (content instanceof StreamSource) {
				InputStreamForker streamForker;
				if (exchangeStreamForked.containsKey(content)) {
					streamForker = exchangeStreamForked.get(messageName);
					streamForker = new InputStreamForker(streamForker
							.getInputStreamTwo());
				} else {
					streamForker = new InputStreamForker(
							((StreamSource) content).getInputStream());
				}
				exchangeStreamForked.put(messageName, streamForker);
				Source source = new StreamSource(streamForker
						.getInputStreamOne());
				messages.get(messageName).setContent(source);
			}
		}
	}
}
