/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ConfigurationDiscovery.java 2:54:36 PM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.configuration;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.NetworkUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * This class is used to discover other Petals containers and set the
 * configuration of this container
 *
 * @author ddesjardins - eBMWebsourcing
 * @author chamerling - eBMWebsourcing
 */
public class ConfigurationService {

    /**
     * Logger
     */
    private LoggingUtil log;

    /**
     * All the already registered containers information
     */
    private List<ContainerInformation> containers;

    /**
     * The initial container information. This information will be modified by
     * the setup method to get an unique container information.
     */
    private ContainerInformation localContainerInformation;

    /**
     * Creates a new instance of {@link ConfigurationService}
     *
     * @param localContainerInformation
     * @param containers
     * @param log
     */
    public ConfigurationService(ContainerInformation localContainerInformation,
        List<ContainerInformation> containers, LoggingUtil log) {
        this.localContainerInformation = localContainerInformation;
        this.containers = containers;
        this.log = log;
    }

    /**
     * Set up the container configuration
     *
     * In distributed mode:
     * <ol>
     * <li>scan the network for other JNDI servers</li>
     * <li>connect to a JNDI server to retrieve the list of
     * ContainerInformation</li>
     * <li>set up the properties of this container (change the ports if they
     * are already taken)</li>
     * <li>write the new configuration on the harddrive</li>
     * </ol>
     */
    public ContainerInformation setUpConfiguration() {
        log.start();
        List<String> ids = new ArrayList<String>();
        List<String> takenPorts = new ArrayList<String>();

        // retrieve the containers informations
        for (ContainerInformation information : containers) {
            if (information.getJoramId() != localContainerInformation.getJoramId()) {
                ids.add(information.getJoramId());
                if (localContainerInformation.getHost().equals(
                    information.getHost())) {
                    takenPorts.add(information.getJndiPort());
                    takenPorts.add(information.getJoramTCPPort());
                    // when PEtALS is embedded in an application server,
                    // the HTML port is not used.
                    if (information.getHtmlPort() != ContainerInformation.UNUSED) {
                        takenPorts.add(information.getHtmlPort());
                    }
                    takenPorts.add(information.getJoramDomainPort());
                    takenPorts.add(information.getDreamTCPPort());
                    takenPorts.add(information.getJmxJndiPort());
                }
            }
        }

        findNotTakenValues(ids, takenPorts);

        persistConfiguration();

        log.end();
        return localContainerInformation;
    }

    /**
     * Find a not taken value in a range
     *
     * @param takens
     *            List of already taken values
     * @param startValue
     *            start value
     * @param range
     *            range of values
     * @param isAPort
     *            <code>true</code> if we search for a port value
     * @return a not taken value in the range from the start value
     */
    protected String findNotTaken(List<String> takens, int startValue,
        int range, boolean isAPort) {
        int i = startValue;
        if (isAPort) {
            while (i < (startValue + range)
                && (takens.contains("" + i) || NetworkUtil
                    .isPortAlreadyUsed(
                        localContainerInformation.getHost(), i))) {
                i++;
            }
        } else {
            while (i < (startValue + range) && takens.contains("" + i)) {
                i++;
            }
        }

        return "" + i;
    }

    /**
     * Find not taken values in the id list and in the port list
     *
     * @param takenIds
     *            taken ids
     * @param takenPorts
     *            taken ports
     */
    private void findNotTakenValues(List<String> takenIds,
        List<String> takenPorts) {
        // Find new id, note that the JNDI port is already resolve by the TRIBe Agent
        
        if (localContainerInformation.getJoramId() == null)  {
            // Find new JORAM Id
            String newId = findNotTaken(takenIds, Integer
                .parseInt(SystemUtil.getJoramId()), 255, false);
            
            if (!newId.equals(SystemUtil.getJoramId())) {
                log.debug("A new id has been set for the container : " + newId);
                SystemUtil.setJoramId(newId);
                if (localContainerInformation.getName() == null) {
                    SystemUtil.setContainerName(newId);
                }
            }
            localContainerInformation.setJoramId(newId);
        }
        
        if (localContainerInformation.getName() == null)  {
            localContainerInformation.setName(localContainerInformation.getJoramId());
        }

        if (localContainerInformation.getJoramTCPPort() == null) {
            // Find new JORAM tcp port
            String newTCPPort = findNotTaken(takenPorts, Integer
                .parseInt(SystemUtil.getJoramTCPPort()), 255,
                true);
            if (!newTCPPort.equals(SystemUtil.getJoramTCPPort())) {
                takenPorts.add(newTCPPort);
                log
                    .debug("A new Joram tcp port has been set for the container : "
                        + newTCPPort);
                SystemUtil.setJoramTCPPort(newTCPPort);
            }
            localContainerInformation.setJoramTCPPort(newTCPPort);
        }

        // Find new JORAM domain port
        if (localContainerInformation.getJoramDomainPort() == null) {
            // Find new JORAM domain port
            String newDomainPort = findNotTaken(takenPorts, Integer
                .parseInt(SystemUtil.getJoramDomainPort()), 255,
                true);
            if (!newDomainPort.equals(SystemUtil.getJoramDomainPort())) {
                takenPorts.add(newDomainPort);
                log.debug("A new Joram domain port has been set for the "
                    + " container : " + newDomainPort);
                SystemUtil.setJoramDomainPort(newDomainPort);
            }
            localContainerInformation.setJoramDomainPort(newDomainPort);
        }

        if (localContainerInformation.getHtmlPort() == null) {
            // Find new html port
            String newHTMLPort = findNotTaken(takenPorts, Integer
                .parseInt(SystemUtil.getHtmlPort()), 255, true);
            if (!newHTMLPort.equals(SystemUtil.getHtmlPort())) {
                takenPorts.add(newHTMLPort);
                log.debug("A new HTML port has been set for the container : "
                    + newHTMLPort);
                SystemUtil.setHtmlPort(newHTMLPort);
            }
            localContainerInformation.setHtmlPort(newHTMLPort);
        }

        if (localContainerInformation.getJmxJndiPort() == null) {
            // Find new jmx jndi port
            String newJMXJndiPort = findNotTaken(takenPorts, Integer
                .parseInt(SystemUtil.getJmxJndiPort()), 255, true);
            if (!newJMXJndiPort.equals(SystemUtil.getJmxJndiPort())) {
                takenPorts.add(newJMXJndiPort);
                log.debug("A new JMX JNDI port has been set for the container : "
                    + newJMXJndiPort);
                SystemUtil.setJmxJndiPort(newJMXJndiPort);
            }
            localContainerInformation.setJmxJndiPort(newJMXJndiPort);
        }
        
        if (localContainerInformation.getDreamTCPPort() == null) {
            // Find new jmx port
            String newDreamTCPPort = findNotTaken(takenPorts, Integer
                .parseInt(SystemUtil.getDreamTCPPort()), 255, true);
            if (!newDreamTCPPort.equals(SystemUtil.getDreamTCPPort())) {
                takenPorts.add(newDreamTCPPort);
                log.debug("A new Dream TCP port has been set for the container : "
                    + newDreamTCPPort);
                SystemUtil.setDreamTCPPort(newDreamTCPPort);
            }
            localContainerInformation.setDreamTCPPort(newDreamTCPPort);
        }
    }

    /**
     * Persists the configuration into a properties file
     *
     */
    private void persistConfiguration() {
        SystemUtil.persist();
    }
}
