/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Monitoring.java msauvage $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.monitoring.router;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.MessageExchange.Role;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.kernel.server.MBeanHelper;

public class Monitoring implements MonitoringMBean, NotificationBroadcaster {

    /**
     * Type of notification sent
     */
    private static final String NEW_MSG_NOTIF_TYPE = "petals.notif.newMsg";

    /**
     * The notifications sequence number
     */
    private static int sequenceNumber = 0;

    protected RouterMonitor routerMonitor;

    /**
     * the notification support
     */
    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport();

    /**
     * Stores the messages waiting to be sent
     */
    private List<Map<String, Object>> messages = new ArrayList<Map<String, Object>>();

    /**
     * Size of list of messages to send by notification. Default value is set to
     * 50
     */
    private int messagesThreshold = 50;

    /**
     * Time in milliseconds to wait before sending messages. Default is 10000
     * (i.e. 10 seconds)
     */
    private int timeSendLimit = 10000;

    /**
     * A thread used to wait a given time before sending new messages, if
     * messages threshold has not been reached
     */
    private MonitoringThread monitoringThread;

    /**
     * Creates a simple, empty Monitoring object
     * 
     */
    public Monitoring() {
        monitoringThread = new MonitoringThread(this);
    }

    /**
     * 
     * @return
     * @throws TransformerConfigurationException
     */
    protected static Transformer getTransformer()
        throws TransformerConfigurationException {
        Transformer transformer = TransformerFactory.newInstance()
            .newTransformer();
        if (transformer == null) {
            transformer = TransformerFactory.newInstance().newTransformer();
            Properties props = new Properties();
            props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperties(props);
        }
        return transformer;
    }

    /**
     * start the monitoring with or without monitoring content of messages. If
     * monitoring is already started, just modify the monitoring of messages
     * content.
     * 
     * @param showMessageContent
     *            indicates if content of messages have to be monitored, true
     *            for monitor the content, false for ignoring it
     */
    public void activateMonitoring(boolean showMessageContent) {
        if (routerMonitor != null) {
            if (!routerMonitor.isMonitoring()) {
                routerMonitor.activateMonitoring(showMessageContent);
                startMonitoringThread();
            } else {
                routerMonitor.showMessageContent(showMessageContent);
            }
        }
    }

    /**
     * @see NotificationBroadcasterSupport#addNotificationListener(NotificationListener,
     *      NotificationFilter, Object)
     */
    public void addNotificationListener(NotificationListener listener,
        NotificationFilter filter, Object handback)
        throws IllegalArgumentException {
        notifSupport.addNotificationListener(listener, filter, handback);
    }

    /**
     * stops the monitoring of messages and their content.
     */
    public void deactivateMonitoring() {
        if (routerMonitor != null) {
            if (routerMonitor.isMonitoring()) {
                routerMonitor.deactivateMonitoring();
            }
            stopMonitoringThread();
        }
    }

    /**
     * Return the threshold number of messages to send
     */
    public int getMessagesThreshold() {
        return messagesThreshold;
    }

    /**
     * "id" = <code>String</code> exchangeId <br>
     * "time" = <code>long</code> capture.time <br>
     * "component" = <code>String</code> capture.component<br>
     * "role" = <code>String</code> capture.role<br>
     * "type" = <code>String</code> capture.messageType <br>
     * "status" = <code>String</code> capture.status <br>
     * "content" = <code>String</code> capture.content <br>
     * "serviceName" = <code>String</code> capture.serviceName <br>
     * "endpointName" = <code>String</code> capture.endpointName <br>
     * 
     * @see NotificationBroadcasterSupport#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types = new String[] {NEW_MSG_NOTIF_TYPE};
        String name = Notification.class.getName();
        String description = "Notification containing a list of new messages : "
            + "HashMap{(String) id, (long) time,(String) component, "
            + "(String) role, (String) type, (String) status, (String) content, "
            + "(String serviceName, (String) endpointName}";
        MBeanNotificationInfo info = new MBeanNotificationInfo(types, name,
            description);
        return new MBeanNotificationInfo[] {info};
    }

    /**
     * Gets the router monitor
     * 
     * @return
     */
    public RouterMonitor getRouterMonitor() {
        return routerMonitor;
    }

    /**
     * Returns the time to wait between two notifications if messages threshold
     * number has not been reached. Returns -1 if monitoring not started.
     * 
     * @return an int representation of the time to wait, -1 if monitoring not
     *         active
     */
    public int getTimeSendLimit() {
        return timeSendLimit;
    }

    /**
     * Indicates if content of messages have to be set in the messages monitored
     * 
     * @return true if content of messages have to be in the messages, false if
     *         not
     */
    public boolean isMessageContentShown() {
        return routerMonitor.isMessageContentShown();
    }

    /**
     * Indicates if monitoring is active or not.
     * 
     * @return true if monitoring is running, false if not.
     */
    public boolean isMonitoring() {
        return routerMonitor.isMonitoring();
    }

    /**
     * Sends notification of all messages stored to all listeners
     */
    public synchronized void notifyMessages() {
        if (messages.size() > 0) {
            try {
                if (monitoringThread != null) {
                    restartMonitoringThread();
                }
                Notification n = new Notification(NEW_MSG_NOTIF_TYPE,
                    new ObjectName("Petals:name="
                        + MBeanHelper.MONITORING_MBEAN + ",type=service"),
                    sequenceNumber++, messages.size() + " new messages coming");
                // notifying a copy of the list to be able to reset the list
                n.setUserData(new ArrayList<Map<String, Object>>(messages));
                notifSupport.sendNotification(n);
                // reseting messages list and thread
                messages.clear();
            } catch (MalformedObjectNameException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * @see NotificationBroadcasterSupport#removeNotificationListener(NotificationListener)
     */
    public void removeNotificationListener(NotificationListener listener)
        throws ListenerNotFoundException {
        notifSupport.removeNotificationListener(listener);
    }

    /**
     * Sets the threshold of messages to send.
     * 
     * @param threshold
     *            the new number of messages
     */
    public void setMessagesThreshold(int threshold) {
        this.messagesThreshold = threshold;
    }

    /**
     * Sets the router monitor
     * 
     * @param routerMonitor
     *            the new router monitor to be set
     */
    public void setRouterMonitor(RouterMonitor routerMonitor) {
        this.routerMonitor = routerMonitor;
    }

    /**
     * Sets the time to wait between two notifications if messages threshold
     * number has not been reached
     * 
     * @param time
     *            the time to wait in milliseconds
     */
    public void setTimeSendLimit(int time) {
        timeSendLimit = time;
        if (monitoringThread != null) {
            monitoringThread.setSendingFrequency(timeSendLimit);
        }
    }

    /**
     * Adds a message to the list of messages stored. If threshold number of
     * messages stored is reached, calls the notifyMessages() method to send the
     * messages via a notification to all listeners
     * 
     */
    protected void addMessage(String exchangeId, MessageExchangeImpl exchange,
        boolean showMessageContent) {
        messages.add(messageToRaw(exchangeId, exchange, showMessageContent));
        if (messages.size() >= messagesThreshold) {
            notifyMessages();
        }
    }

    /**
     * Gets the monitoring thread
     * 
     * @return a MonitorThread
     */
    protected MonitoringThread getMonitoringThread() {
        return monitoringThread;
    }

    /**
     * Restarts the monitoring thread by destroying the one existing and then
     * creating and starting a new one
     * 
     */
    protected void restartMonitoringThread() {
        if (monitoringThread!=null)
        {
            monitoringThread.setSkipNextWakeUp(true);
        }
    }

    /**
     * Sets the monitoring thread to a thread given
     * 
     * @param monitoringThread
     *            the new MonitoringThread to use
     */
    protected void setMonitoringThread(MonitoringThread monitoringThread) {
        this.monitoringThread = monitoringThread;
    }

    /**
     * Creates a new MonitoringThread if no one is set and starts it
     * 
     */
    private void startMonitoringThread() {
        if (monitoringThread == null) {
            monitoringThread = new MonitoringThread(this);
        }
        monitoringThread.start();
    }

    /**
     * Stops and destroys the active monitoring thread
     * 
     */
    private void stopMonitoringThread() {
        if (monitoringThread != null) {
            monitoringThread.interrupt();
            monitoringThread = null;
        }
    }

    /**
     * transform a capture to a raw Map : <br />
     * "id" = <code>String</code> exchangeId <br>
     * "time" = <code>long</code> capture.time <br>
     * "component" = <code>String</code> capture.component<br>
     * "role" = <code>String</code> capture.role<br>
     * "type" = <code>String</code> capture.messageType <br>
     * "status" = <code>String</code> capture.status <br>
     * "content" = <code>String</code> capture.content <br>
     * "serviceName" = <code>String</code> capture.serviceName <br>
     * "endpointName" = <code>String</code> capture.endpointName <br>
     * 
     * @param capture
     * @return
     */
    private Map<String, Object> captureToRaw(String exchangeId,
        ExchangeStateCapture capture) {

        Map<String, Object> rawCapture = new HashMap<String, Object>();
        rawCapture.put("id", exchangeId);
        rawCapture.put("time", capture.time);
        rawCapture.put("component", capture.component);
        rawCapture.put("role", roleToRaw(capture.role));
        rawCapture.put("type", capture.messageType);
        rawCapture.put("status", statusToRaw(capture.status));
        rawCapture.put("content", capture.content);
        rawCapture.put("serviceName", capture.serviceName);
        rawCapture.put("endpointName", capture.endpointName);
        rawCapture.put("container", capture.container);

        return rawCapture;
    }

    /**
     * Create a Source from the String. The string must be in UTF-8 format,
     * otherwise, the Source is not created.
     * 
     * @param msg
     *            an UTF-8 String
     * @return the Source built from the String, null if not an UTF-8
     */
    private Source createSource(final String msg) throws IOException {
        StreamSource source = new StreamSource();
        byte[] msgByte = msg.getBytes("UTF-8");
        ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
        source.setInputStream(in);
        return source;
    }

    /**
     * Create a String from a Source. The String will be in UTF-8.
     * 
     * @param s
     * @return
     * @throws Exception
     */
    private String createString(final Source s) throws Exception {

        String result = null;
        try {
        	if (s instanceof StreamSource
					&& ((StreamSource) s).getInputStream()
							.markSupported()) {
                StreamSource ss = (StreamSource) s;
                ss.getInputStream().reset();
                byte[] buf = new byte[ss.getInputStream().available()];
                BufferedInputStream bis = new BufferedInputStream(ss
                    .getInputStream());
                bis.read(buf);
                result = new String(buf);
            } else {
                StringWriter buffer = new StringWriter();
                Result sresult = new StreamResult(buffer);
                Transformer transform = getTransformer();
                transform.transform(s, sresult);
                result = buffer.toString();
            }
        } catch (IOException e) {
            throw new Exception(e);
        } catch (TransformerConfigurationException e) {
            throw new Exception(e);
        } catch (TransformerException e) {
            throw new Exception(e);
        }
        return result;
    }

    /**
     * Transforms a given MessageExchange into an ExchangeStateCapture
     * 
     * @param exchange
     *            the MessageExchangeImpl to convert
     * @param showMessageContent
     *            true if content must be set in the capture, false if not
     * @return
     */
    private ExchangeStateCapture messageToCapture(MessageExchangeImpl exchange,
        boolean showMessageContent) {
        ExchangeStateCapture state = new ExchangeStateCapture();
        state.time = System.currentTimeMillis();
        state.status = exchange.getStatus();
        state.role = exchange.getRole();
        if (exchange.getEndpoint() != null) {
            state.serviceName = exchange.getEndpoint().getServiceName().toString();
            state.endpointName = exchange.getEndpoint().getEndpointName();
        } else if (exchange.getService() != null) {
            state.serviceName = exchange.getService().toString();
            state.endpointName = "";
        } else {
            state.serviceName = "";
            state.endpointName = "";
        }

        if (Role.CONSUMER.equals(state.role)) {
            state.component = exchange.getConsumerEndpoint().getComponentName();
            state.container = exchange.getConsumerEndpoint().getContainerName();
        } else {
            if (exchange.getEndpoint() != null) {
                state.component = ((AbstractEndpoint) exchange.getEndpoint())
                    .getComponentName();
                state.container = ((AbstractEndpoint) exchange.getEndpoint())
                    .getContainerName();
            } else {
                state.component = "";
                state.container = "";
            }
        }
        if (exchange.getFault() != null) {
            Fault fault = exchange.getFault();
            state.messageType = "fault";
            if (showMessageContent) {
                if (fault.getContent() != null) {
                    // FIXME ugly way to copy the fault content
                    try {
                        state.content = createString(fault.getContent());
                    } catch (Exception e) {
                        state.content = "<monitoring-error>Error occured "
                            + "while getting fault content</monitoring-error>";
                    }
                    try {
                        fault.setContent(createSource(state.content));
                    } catch (MessagingException e) {
                        e.printStackTrace();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        } else {
            Map<String, NormalizedMessage> messages = exchange.getMessages();
            if (messages != null && messages.size() > 0) {
                NormalizedMessage message;
                // searching for an "out" message
                // if not found -> taking first message
                if (messages.get("out") != null) {
                    message = messages.get("out");
                    state.messageType = "out";
                } else {
                    state.messageType = messages.keySet().iterator().next();
                    message = messages.get(state.messageType);
                }
                // adding message content and fault if wanted
                if (showMessageContent) {
                    if (message.getContent() != null) {
                        // FIXME ugly way to copy the message content
                        try {
                            state.content = createString(message.getContent());
                        } catch (Exception e) {
                            state.content = "<monitoring-error>Error occured "
                                + "while getting message content</monitoring-error>";
                        }
                        try {
                            message.setContent(createSource(state.content));
                        } catch (MessagingException e) {
                            e.printStackTrace();
                        } catch (IOException e) {
                            e.printStackTrace();
                        }
                    }
                }
            }
        }
        return state;
    }

    /**
     * Transform a MessageExchange to a raw format :
     * <code>Map<String,Object></code> a raw map representation of the
     * message
     * 
     * @param exchangeId
     *            the id of the exchange containing the message
     * @param exchange
     *            the message exchange
     * @param showMessageContent
     *            true if content must be set in the message, false if not
     * @return a Map<String,Object> representation of the MessageExchange
     */
    private Map<String, Object> messageToRaw(String exchangeId,
        MessageExchangeImpl exchange, boolean showMessageContent) {
        return captureToRaw(exchangeId, messageToCapture(exchange,
            showMessageContent));
    }

    /**
     * Convert a Role to a string representation
     * 
     * @param role
     * @return "consumer" or "provider
     */
    private String roleToRaw(Role role) {
        return (Role.CONSUMER.equals(role)) ? "consumer" : "provider";
    }

    /**
     * Convert a status to a string representation
     * 
     * @param status
     * @return "active" or "done" or "error"
     */
    private String statusToRaw(ExchangeStatus status) {
        return (ExchangeStatus.DONE.equals(status)) ? "done"
            : (ExchangeStatus.ACTIVE.equals(status)) ? "active" : "error";
    }

}
