/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009 OW2, http://www.ow2.org/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.extensions.transporter.nio.util.stream.queue;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * <p>
 * This class implements an output stream in which the data are written into a
 * byte array queue. Each write creates a new entry in the queue. The new entry
 * is a copy of the buffer to write.
 * </p>
 * <p>
 * This class is thread-safe, and should be used with {@link QueueInputStream}.
 * </p>
 * <p>
 * Closing a <tt>QueueOutputStream</tt> has no effect. For performance reasons
 * (no check is done), the methods in this class can be called after the stream
 * has been closed without generating an <tt>IOException</tt>.
 * </p>
 * 
 * @author Christophe DENEUX - Capgemini Sud
 * 
 */

public class QueueOutputStream extends OutputStream {

    protected final Queue<byte[]> bufferList = new ConcurrentLinkedQueue<byte[]>();

    /**
	 * Writes the specified byte to this byte array linked list output stream.
	 * 
	 * @param b
	 *            the byte to be written.
	 * 
	 * @see {@link OutputStream#write(int)}
	 */
    @Override
    public void write(int b) throws IOException {

        final byte newbuf[] = new byte[1];
        newbuf[0] = (byte) b;

        this.write(newbuf);

    }

    /**
	 * <p>
	 * Writes <code>len</code> bytes from the specified byte array starting at
	 * offset <code>off</code> to this byte array queue output stream.
	 * </p>
	 * <p>
	 * Note:
	 * <ol>
	 * <li>this method is overriden to increase performance (we create only one
	 * buffer in the queue, instead of one per byte),</li>
	 * <li>
	 * if the provided buffer must be completely written (
	 * <code>off == 0 &amp;&amp; len == b.length</code>), it is added by
	 * reference in the queue,</li>
	 * <li>if the provided buffer must be partially written, the part to write
	 * is cloned.</li>
	 * </ol>
	 * </p>
	 * 
	 * @param b
	 *            the data buffer to writer.
	 * @param off
	 *            the start offset in the data.
	 * @param len
	 *            the number of bytes to write.
	 */
    @Override
    public void write(byte b[], int off, int len) {
        if ((off < 0) || (off > b.length) || (len < 0) || ((off + len) > b.length)
                || ((off + len) < 0)) {
            throw new IndexOutOfBoundsException();
        } else if (len == 0) {
        	// No byte to write
            return;
        } else if (off == 0 && len == b.length) {
        	// The buffer must be completely written
        	this.bufferList.add(b);
        } else {
        	// Only a part of the buffer must be written
        	final byte newbuf[] = new byte[len];
            System.arraycopy(b, off, newbuf, 0, len);

            this.bufferList.add(newbuf);
        }
    }

}
